# Copyright (c)      2013 - Christophe-Marie Duquesne
# Copyright (c) 2013-2023 - Simon Conseil
# Copyright (c)      2021 - Keith Feldman

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import logging
import os
import re
import shutil
import subprocess
from os.path import splitext

from PIL import Image as PILImage

from . import image, utils
from .utils import is_valid_html5_video


class SubprocessException(Exception):
    pass


def check_subprocess(cmd, source, outname=None):
    """Run the command to resize the video and remove the output file if the
    processing fails.

    """
    logger = logging.getLogger(__name__)
    try:
        res = subprocess.run(cmd, capture_output=True)
    except KeyboardInterrupt:
        logger.debug("Process terminated, removing file %s", outname)
        if outname and os.path.isfile(outname):
            os.remove(outname)
        raise

    if res.returncode:
        logger.debug("STDOUT:\n %s", res.stdout.decode("utf8"))
        logger.debug("STDERR:\n %s", res.stderr.decode("utf8"))
        if outname and os.path.isfile(outname):
            logger.debug("Removing file %s", outname)
            os.remove(outname)
        raise SubprocessException("Failed to process " + source)


def video_size(source, converter="ffmpeg"):
    """Return the dimensions of the video."""
    res = subprocess.run([converter, "-i", source], stderr=subprocess.PIPE)
    stderr = res.stderr.decode("utf8", errors="ignore")
    pattern = re.compile(r"Stream.*Video.* ([0-9]+)x([0-9]+)")
    match = pattern.search(stderr)
    rot_pattern = re.compile(r"rotate\s*:\s*-?(90|270)")
    rot_match = rot_pattern.search(stderr)

    if match:
        x, y = int(match.groups()[0]), int(match.groups()[1])
    else:
        x = y = 0
    if rot_match:
        x, y = y, x
    return x, y


def get_resize_options(source, converter, output_size):
    """Figure out resize options for video from src and dst sizes.

    :param source: path to a video
    :param outname: path to the generated video
    :param settings: settings dict
    """
    logger = logging.getLogger(__name__)
    w_src, h_src = video_size(source, converter=converter)
    w_dst, h_dst = output_size
    logger.debug("Video size: %i, %i -> %i, %i", w_src, h_src, w_dst, h_dst)

    # do not resize if input dimensions are smaller than output dimensions
    if w_src <= w_dst and h_src <= h_dst:
        return []

    # http://stackoverflow.com/questions/8218363/maintaining-ffmpeg-aspect-ratio
    # + I made a drawing on paper to figure this out
    if h_dst * w_src < h_src * w_dst:
        # biggest fitting dimension is height
        resize_opt = ["-vf", "scale=trunc(oh*a/2)*2:%i" % h_dst]
    else:
        # biggest fitting dimension is width
        resize_opt = ["-vf", "scale=%i:trunc(ow/a/2)*2" % w_dst]

    return resize_opt


def _get_empty_if_none_else_variable(variable):
    return [] if not variable else variable


def generate_video_pass(converter, source, options, outname=None):
    """Run a single pass of encoding.

    :param source: source video
    :param options: options to pass to encoder
    :param outname: if multi-pass, this is None on the first pass
    """
    logger = logging.getLogger(__name__)
    outname_opt = [] if not outname else [outname]
    # Encoding options improved, thanks to
    # http://ffmpeg.org/trac/ffmpeg/wiki/vpxEncodingGuide
    cmd = [converter, "-i", source, "-y"]  # -y to overwrite output files
    cmd += options + outname_opt
    logger.debug("Processing video: %s", " ".join(cmd))
    check_subprocess(cmd, source, outname=outname)


def generate_video(source, outname, settings):
    """Video processor.

    :param source: path to a video
    :param outname: path to the generated video
    :param settings: settings dict
    :param options: array of options passed to ffmpeg

    """
    logger = logging.getLogger(__name__)

    video_format = settings.get("video_format")
    options = settings.get(video_format + "_options")
    second_pass_options = settings.get(video_format + "_options_second_pass")
    video_always_convert = settings.get("video_always_convert")
    converter = settings["video_converter"]

    resize_opt = []
    if settings.get("video_size"):
        resize_opt = get_resize_options(source, converter, settings["video_size"])

    base, src_ext = splitext(source)
    base, dst_ext = splitext(outname)

    if dst_ext == src_ext and not resize_opt and not video_always_convert:
        logger.debug(
            "For %s, the source and destination extension are the "
            "same, there is no resizing to be done, and "
            "video_always_convert is False, so the output is "
            " being copied",
            outname,
        )
        shutil.copy(source, outname)
        return

    final_pass_options = _get_empty_if_none_else_variable(options) + resize_opt
    if second_pass_options:
        generate_video_pass(converter, source, final_pass_options)
        final_second_pass_options = (
            _get_empty_if_none_else_variable(second_pass_options) + resize_opt
        )
        generate_video_pass(converter, source, final_second_pass_options, outname)
    else:
        generate_video_pass(converter, source, final_pass_options, outname)


def generate_thumbnail(
    source,
    outname,
    box,
    delay,
    fit=True,
    options=None,
    converter="ffmpeg",
    black_retries=0,
    black_offset=1,
    black_max_colors=4,
):
    """Create a thumbnail image for the video source, based on ffmpeg."""
    logger = logging.getLogger(__name__)
    tmpfile = outname + ".tmp.jpg"

    currentTry = 0
    iDelay = int(delay)
    while currentTry <= abs(black_retries):
        # dump an image of the video
        cmd = [converter, "-i", source, "-an", "-r", "1"]
        cmd += ["-ss", str(iDelay), "-vframes", "1", "-y", tmpfile]
        logger.debug("Create thumbnail for video: %s", " ".join(cmd))
        check_subprocess(cmd, source, outname)
        if os.path.isfile(tmpfile) and black_retries > 0:
            with PILImage.open(tmpfile) as img:
                colors = img.getcolors(maxcolors=black_max_colors)
            if colors is None:
                # There were more colors than maxcolors in the image, it
                # looks suitable for a valid thumbnail
                break
            else:
                # Only found 'maxcolors' unique colors, looks like a solid
                # color, try again with another seek delay
                currentTry += 1
                iDelay += abs(black_offset)
        else:
            break

    # Sometimes ffmpeg fails with returncode zero but without producing an
    # output file Thus, we need to check if an output file was created. If
    # not, assume ffmpeg failed
    if not os.path.isfile(tmpfile):
        logger.debug("Thumbnail generation failed. Likely due to short video length.")
        cmd = [converter, "-i", source, "-an", "-r", "1"]
        cmd += ["-ss", "0", "-vframes", "1", "-y", tmpfile]
        logger.debug("Retry to create thumbnail for video: %s", " ".join(cmd))
        check_subprocess(cmd, source, outname)

    # use the generate_thumbnail function from sigal.image
    image.generate_thumbnail(tmpfile, outname, box, fit=fit, options=options)
    # remove the image
    os.unlink(tmpfile)


def process_video(media):
    """Process a video: resize, create thumbnail."""
    logger = logging.getLogger(__name__)
    settings = media.settings

    with utils.raise_if_debug() as status:
        if settings["use_orig"] and is_valid_html5_video(media.src_ext):
            utils.copy(media.src_path, media.dst_path, symlink=settings["orig_link"])
        else:
            valid_formats = ["mp4", "webm"]
            video_format = settings["video_format"]

            if video_format not in valid_formats:
                logger.error(
                    "Invalid video_format. Please choose one of: %s", valid_formats
                )
                raise ValueError
            generate_video(media.src_path, media.dst_path, settings)

        if settings["make_thumbs"]:
            generate_thumbnail(
                media.dst_path,
                media.thumb_path,
                settings["thumb_size"],
                settings["thumb_video_delay"],
                fit=settings["thumb_fit"],
                options=settings["jpg_options"],
                converter=settings["video_converter"],
                black_retries=settings["thumb_video_black_retries"],
                black_offset=settings["thumb_video_black_retry_offset"],
                black_max_colors=settings["thumb_video_black_max_colors"],
            )

    return status.value



"""Plugin to compress static files for faster HTTP transfer.

Your web server must be configured in order to select pre-compressed
files instead of the required one, to save CPU. For example, in nginx::

    server {
        gzip_static;
    }

Currently, 3 methods are supported:

- ``gzip``: No dependency required.
  This is the fastest, but also largest output.
- ``zopfli``: Requires zopfli_.
  gzip compatible output with optimized size.
- ``brotli``: Requires brotli_.
  Brotli is the best compressor for web usage.

.. _zopfli: https://pypi.org/project/zopfli/
.. _brotli: https://pypi.org/project/Brotli/

Settings available as dictionary in ``compress_assets_options``:

- ``method``: One of ``gzip``, ``zopfli`` or ``brotli``.
- ``suffixes``: List of file suffixes eligible to compression.
  Default to ``['htm', 'html', 'css', 'js', 'svg']``

"""

import gzip
import logging
import os
import shutil

from click import progressbar

from sigal import signals

logger = logging.getLogger(__name__)

DEFAULT_SETTINGS = {
    "suffixes": ["htm", "html", "css", "js", "svg"],
    "method": "gzip",
}


class BaseCompressor:
    suffix = None

    def __init__(self, settings):
        self.suffixes_to_compress = settings.get(
            "suffixes", DEFAULT_SETTINGS["suffixes"]
        )

    def do_compress(self, filename, compressed_filename):
        """
        Perform actual compression.
        This should be implemented by subclasses.
        """
        raise NotImplementedError

    def compress(self, filename):
        """Compress a file, only if needed."""
        compressed_filename = self.get_compressed_filename(filename)
        if not compressed_filename:
            return

        self.do_compress(filename, compressed_filename)

    def get_compressed_filename(self, filename):
        """If the given filename should be compressed, returns the
        compressed filename.

        A file can be compressed if:

        - It is a whitelisted extension
        - The compressed file does not exist
        - The compressed file exists by is older than the file itself

        Otherwise, it returns False.

        """
        if os.path.splitext(filename)[1][1:] not in self.suffixes_to_compress:
            return False

        file_stats = None
        compressed_stats = None
        compressed_filename = f"{filename}.{self.suffix}"
        try:
            file_stats = os.stat(filename)
            compressed_stats = os.stat(compressed_filename)
        except OSError:  # FileNotFoundError is for Python3 only
            pass

        if file_stats and compressed_stats:
            return (
                compressed_filename
                if file_stats.st_mtime > compressed_stats.st_mtime
                else False
            )
        else:
            return compressed_filename


class GZipCompressor(BaseCompressor):
    suffix = "gz"

    def do_compress(self, filename, compressed_filename):
        with open(filename, "rb") as f_in, gzip.open(
            compressed_filename, "wb"
        ) as f_out:
            shutil.copyfileobj(f_in, f_out)


class ZopfliCompressor(BaseCompressor):
    suffix = "gz"

    def do_compress(self, filename, compressed_filename):
        import zopfli.gzip

        with open(filename, "rb") as f_in, open(compressed_filename, "wb") as f_out:
            f_out.write(zopfli.gzip.compress(f_in.read()))


class BrotliCompressor(BaseCompressor):
    suffix = "br"

    def do_compress(self, filename, compressed_filename):
        import brotli

        with open(filename, "rb") as f_in, open(compressed_filename, "wb") as f_out:
            f_out.write(brotli.compress(f_in.read(), mode=brotli.MODE_TEXT))


def get_compressor(settings):
    name = settings.get("method", DEFAULT_SETTINGS["method"])
    if name == "gzip":
        return GZipCompressor(settings)
    elif name == "zopfli":
        try:
            import zopfli.gzip  # noqa

            return ZopfliCompressor(settings)
        except ImportError:
            logging.error("Unable to import zopfli module")

    elif name == "brotli":
        try:
            import brotli  # noqa

            return BrotliCompressor(settings)
        except ImportError:
            logger.error("Unable to import brotli module")

    else:
        logger.error(f"No such compressor {name}")


def compress_gallery(gallery):
    logging.info("Compressing assets for %s", gallery.title)
    compress_settings = gallery.settings.get(
        "compress_assets_options", DEFAULT_SETTINGS
    )
    compressor = get_compressor(compress_settings)

    if compressor is None:
        return

    # Collecting theme assets
    theme_assets = []
    for current_directory, _, filenames in os.walk(
        os.path.join(gallery.settings["destination"], "static")
    ):
        for filename in filenames:
            theme_assets.append(os.path.join(current_directory, filename))

    with progressbar(
        length=len(gallery.albums) + len(theme_assets), label="Compressing static files"
    ) as bar:
        for album in gallery.albums.values():
            compressor.compress(os.path.join(album.dst_path, album.output_file))
            bar.update(1)

        for theme_asset in theme_assets:
            compressor.compress(theme_asset)
            bar.update(1)


def register(settings):
    if settings["write_html"]:
        signals.gallery_build.connect(compress_gallery)



# copyright (c) 2020 Bowen Ding

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import logging
import os
import pickle
import random
import string
from io import BytesIO
from itertools import chain

from click import progressbar

from sigal import signals
from sigal.settings import get_thumb
from sigal.utils import copy

from .endec import encrypt, kdf_gen_key

logger = logging.getLogger(__name__)

ASSETS_PATH = os.path.normpath(
    os.path.join(os.path.abspath(os.path.dirname(__file__)), "static")
)


class Abort(Exception):
    pass


def gen_rand_string(length=16):
    return "".join(
        random.SystemRandom().choices(string.ascii_letters + string.digits, k=length)
    )


def get_options(settings, cache):
    if "encrypt_options" not in settings:
        logging.error("Encrypt: no encrypt_options in settings")
        raise ValueError("no encrypt_options in settings")

    # try load credential from cache
    try:
        options = cache["credentials"]
    except KeyError:
        options = settings["encrypt_options"]

    table = str.maketrans({'"': r"\"", "\\": r"\\"})
    if (
        "password" not in settings["encrypt_options"]
        or len(settings["encrypt_options"]["password"]) == 0
    ):
        logger.error("Encrypt: no password provided")
        raise ValueError("no password provided")
    else:
        options["password"] = settings["encrypt_options"]["password"]
        options["escaped_password"] = options["password"].translate(table)

    if "ask_password" not in options:
        options["ask_password"] = settings["encrypt_options"].get("ask_password", False)
    options["filtered_password"] = (
        "" if options["ask_password"] else options["escaped_password"]
    )

    if "gcm_tag" not in options:
        options["gcm_tag"] = gen_rand_string()
    options["escaped_gcm_tag"] = options["gcm_tag"].translate(table)

    if "kdf_salt" not in options:
        options["kdf_salt"] = gen_rand_string()
    options["escaped_kdf_salt"] = options["kdf_salt"].translate(table)

    if "galleryId" not in options:
        options["galleryId"] = gen_rand_string(6)

    if "kdf_iters" not in options:
        options["kdf_iters"] = 10000

    # in case any of the credentials are newly generated, write them back
    # to cache
    cache["credentials"] = {
        "gcm_tag": options["gcm_tag"],
        "kdf_salt": options["kdf_salt"],
        "kdf_iters": options["kdf_iters"],
        "galleryId": options["galleryId"],
    }

    return options


def cache_key(media):
    return os.path.join(media.path, media.dst_filename)


def save_property(cache, media):
    key = cache_key(media)
    if key not in cache:
        cache[key] = {}
        cache[key]["size"] = media.size
        cache[key]["thumb_size"] = media.thumb_size
        cache[key]["encrypted"] = set()


def get_encrypt_list(settings, media):
    to_encrypt = []
    # resized image or in case of "use_orig", the original
    to_encrypt.append(media.dst_filename)
    if settings["make_thumbs"]:
        to_encrypt.append(get_thumb(settings, media.dst_filename))  # thumbnail
    if media.big is not None and not settings["use_orig"]:
        to_encrypt.append(media.big)  # original image
    to_encrypt = list(map(lambda path: os.path.join(media.path, path), to_encrypt))
    return to_encrypt


def load_property(album):
    gallery = album.gallery
    cache = load_cache(gallery.settings)

    for media in album.medias:
        if media.type == "image":
            key = cache_key(media)
            if key in cache:
                media.size = cache[key]["size"]
                media.thumb_size = cache[key]["thumb_size"]


def load_cache(settings):
    cachePath = os.path.join(settings["destination"], ".encryptCache")
    try:
        with open(cachePath, "rb") as cacheFile:
            encryptCache = pickle.load(cacheFile)
            logger.debug("Loaded encryption cache with %d entries", len(encryptCache))
            return encryptCache
    except FileNotFoundError:
        encryptCache = {}
        return encryptCache
    except Exception as e:
        logger.error("Could not load encryption cache: %s", e)
        logger.error(
            "Giving up encryption. You may have to delete and "
            "rebuild the entire gallery."
        )
        raise Abort


def save_cache(settings, cache):
    cachePath = os.path.join(settings["destination"], ".encryptCache")
    try:
        with open(cachePath, "wb") as cacheFile:
            pickle.dump(cache, cacheFile)
            logger.debug("Stored encryption cache with %d entries", len(cache))
    except Exception as e:
        logger.warning("Could not store encryption cache: %s", e)
        logger.warning("Next build of the gallery is likely to fail!")


def encrypt_gallery(gallery):
    albums = gallery.albums
    settings = gallery.settings

    cache = load_cache(settings)
    config = get_options(settings, cache)
    logger.debug("encryption config: %s", config)

    logger.info("starting encryption")
    copy_assets(settings)
    encrypt_files(settings, config, cache, albums, gallery.progressbar_target)
    save_cache(settings, cache)


def encrypt_files(settings, config, cache, albums, progressbar_target):
    if settings["keep_orig"] and settings["orig_link"]:
        logger.warning(
            "Original images are symlinked! Encryption is aborted. "
            "Please set 'orig_link' to False and restart gallery build."
        )
        raise Abort

    key = kdf_gen_key(config["password"], config["kdf_salt"], config["kdf_iters"])
    gcm_tag = config["gcm_tag"].encode("utf-8")

    medias = list(chain.from_iterable(albums.values()))
    with progressbar(
        medias,
        label="%16s" % "Encrypting files",
        file=progressbar_target,
        show_eta=True,
    ) as medias:
        for media in medias:
            if media.type != "image":
                logger.info("Skipping non-image file %s", media.src_filename)
                continue

            save_property(cache, media)
            to_encrypt = get_encrypt_list(settings, media)

            cacheEntry = cache[cache_key(media)]["encrypted"]
            for f in to_encrypt:
                if f in cacheEntry:
                    logger.info("Skipping %s as it is already encrypted", f)
                    continue

                full_path = os.path.join(settings["destination"], f)
                if encrypt_file(f, full_path, key, gcm_tag):
                    cacheEntry.add(f)
                else:
                    # save the progress and abort the build if any image
                    # fails to be encrypted
                    save_cache(settings, cache)
                    raise Abort

    key_check_path = os.path.join(settings["destination"], "static", "keycheck.txt")
    encrypt_file("keycheck.txt", key_check_path, key, gcm_tag)


def encrypt_file(filename, full_path, key, gcm_tag):
    with BytesIO() as outBuffer:
        try:
            with open(full_path, "rb") as infile:
                encrypt(key, infile, outBuffer, gcm_tag)
        except Exception as e:
            logger.error("Encryption failed for %s: %s", filename, e)
            return False
        else:
            logger.info("Encrypting %s...", filename)
            try:
                with open(full_path, "wb") as outfile:
                    outfile.write(outBuffer.getbuffer())
            except Exception as e:
                logger.error("Could not write to file %s: %s", filename, e)
                return False
    return True


def copy_assets(settings):
    theme_path = os.path.join(settings["destination"], "static")
    copy(
        os.path.join(ASSETS_PATH, "decrypt.js"),
        theme_path,
        symlink=False,
        rellink=False,
    )
    copy(
        os.path.join(ASSETS_PATH, "keycheck.txt"),
        theme_path,
        symlink=False,
        rellink=False,
    )
    copy(
        os.path.join(ASSETS_PATH, "sw.js"),
        settings["destination"],
        symlink=False,
        rellink=False,
    )


def inject_scripts(context):
    cache = load_cache(context["settings"])
    context["encrypt_options"] = get_options(context["settings"], cache)


def register(settings):
    signals.gallery_build.connect(encrypt_gallery)
    signals.album_initialized.connect(load_property)
    signals.before_render.connect(inject_scripts)



# Copyright 2017 - Tilman 't.animal' Adler

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

""" This plugin offers more fine-grained control over exluded images and
folders, similarly to how it's handled on Android.

To ignore a folder or image put a ``.nomedia`` file next to it in its parent
folder and put its name into the file.  E.g.::

    content of folder:
        IMG_3425.JPG, IMG_2426.JPG, IMG_2427.JPG, subfolder, .nomedia
    content of .nomedia:
        IMG_2426.JPG
        IMG_2427.JPG
        subfolder

will ignore all images but IMG_3425.JPG and the subfolder.

Alternatively, if you put a ``.nomedia`` file into a folder and leave it blank
(i.e. an empty file called ``.nomedia`` in a folder containing images), this
ignores the whole folder it's located in (like on Android).

WARNING: When you have a pre-existing gallery from a previous run of sigal
adding a new ``.nomedia`` file will not remove the newly ignored images/albums
from the existing gallery (only the entries in the parent gallery pointing to
it).  They might still be reachable thereafter. Either remove the whole gallery
to be sure or remove the ignored files/folders inside the gallery to remove
them for good.

"""

import logging
import os

from sigal import signals

logger = logging.getLogger(__name__)


def _remove_albums_with_subdirs(albums, keystoremove, prefix=""):
    for keytoremove in keystoremove:
        for key in list(albums.keys()):
            if key.startswith(prefix + keytoremove):
                # subdirs' target directories have already been created,
                # remove them first
                try:
                    album = albums[key]
                    settings = album.settings
                    if album.medias:
                        os.rmdir(os.path.join(album.dst_path, settings["thumb_dir"]))

                    if album.medias and settings["keep_orig"]:
                        os.rmdir(os.path.join(album.dst_path, settings["orig_dir"]))

                    os.rmdir(album.dst_path)
                except OSError:
                    # directory was created and populated with images in a
                    # previous run => keep it
                    pass

                # now remove the album from the surrounding album/gallery
                del albums[key]


def filter_nomedia(album, settings=None):
    """Removes all filtered Media and subdirs from an Album"""
    nomediapath = os.path.join(album.src_path, ".nomedia")

    if not os.path.isfile(nomediapath):
        return

    if os.path.getsize(nomediapath) == 0:
        logger.info(
            "Ignoring album '%s' because of present 0-byte .nomedia file", album.name
        )

        # subdirs have been added to the gallery already, remove them
        # there, too
        _remove_albums_with_subdirs(album.gallery.albums, [album.path])
        try:
            os.rmdir(album.dst_path)
        except OSError:
            # directory was created and populated with images in a
            # previous run => keep it
            pass

        # cannot set albums => empty subdirs so that no albums are
        # generated
        album.subdirs = []
        album.medias = []

    else:
        with open(nomediapath) as nomediaFile:
            logger.info("Found a .nomedia file in %s, ignoring its entries", album.name)
            ignored = nomediaFile.read().split("\n")

            album.medias = [
                media for media in album.medias if media.src_filename not in ignored
            ]
            album.subdirs = [
                dirname for dirname in album.subdirs if dirname not in ignored
            ]

            # subdirs have been added to the gallery already, remove
            # them there, too
            _remove_albums_with_subdirs(
                album.gallery.albums, ignored, album.path + os.path.sep
            )


def register(settings):
    signals.album_initialized.connect(filter_nomedia)



# Copyright 2017 - Tilman 't.animal' Adler

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

""" Decreases the time needed to build large galleries (e.g.: 25k images in
2.5s instead of 30s)

This plugin allows extended caching, which is useful for large galleries. Once
a gallery has been built it caches all metadata for all media (markdown, exif,
itpc) in the gallery target folder. Before the next run it restores them so
that the image and metadata files do not have to be parsed again. For large
galleries this can speed up the creation of index files dramatically.
"""

import logging
import os
import pickle

from .. import signals
from ..utils import get_mod_date

logger = logging.getLogger(__name__)


def load_metadata(album):
    """Loads the metadata of all media in an album from cache"""
    if not hasattr(album.gallery, "metadataCache"):
        _restore_cache(album.gallery)
    cache = album.gallery.metadataCache

    # load album metadata
    key = os.path.join(album.path, "_index")
    if key in cache:
        data = cache[key]

        # check if file has changed
        try:
            mod_date = int(get_mod_date(album.markdown_metadata_filepath))
        except FileNotFoundError:
            pass
        else:
            if data.get("mod_date", -1) >= mod_date:
                # cache is good
                if "markdown_metadata" in data:
                    album.markdown_metadata = data["markdown_metadata"]

    # load media metadata
    for media in album.medias:
        key = os.path.join(media.path, media.dst_filename)
        if key in cache:
            data = cache[key]

            # check if files have changed
            try:
                mod_date = int(get_mod_date(media.src_path))
            except FileNotFoundError:
                continue
            if data.get("mod_date", -1) < mod_date:
                continue  # file_metadata needs updating

            if "file_metadata" in data:
                media.file_metadata = data["file_metadata"]
            if "exif" in data:
                media.exif = data["exif"]

            try:
                mod_date = int(get_mod_date(media.markdown_metadata_filepath))
            except FileNotFoundError:
                continue
            if data.get("meta_mod_date", -1) < mod_date:
                continue  # markdown_metadata needs updating

            if "markdown_metadata" in data:
                media.markdown_metadata = data["markdown_metadata"]


def _restore_cache(gallery):
    """Restores the metadata cache from the cache file"""
    cachePath = os.path.join(gallery.settings["destination"], ".metadata_cache")
    try:
        if os.path.exists(cachePath):
            with open(cachePath, "rb") as cacheFile:
                gallery.metadataCache = pickle.load(cacheFile)
                logger.debug("Loaded cache with %d entries", len(gallery.metadataCache))
        else:
            gallery.metadataCache = {}
    except Exception as e:
        logger.warning("Could not load cache: %s", e)
        gallery.metadataCache = {}


def save_cache(gallery):
    """Stores the exif data of all images in the gallery"""

    if hasattr(gallery, "metadataCache"):
        cache = gallery.metadataCache
    else:
        cache = gallery.metadataCache = {}

    for album in gallery.albums.values():
        try:
            data = {
                "mod_date": int(get_mod_date(album.markdown_metadata_filepath)),
                "markdown_metadata": album.markdown_metadata,
            }
            cache[os.path.join(album.path, "_index")] = data
        except FileNotFoundError:
            pass

        for media in album.medias:
            data = {}
            try:
                mod_date = int(get_mod_date(media.src_path))
            except FileNotFoundError:
                continue
            else:
                data["mod_date"] = mod_date
                data["file_metadata"] = media.file_metadata
                if hasattr(media, "exif"):
                    data["exif"] = media.exif

            try:
                meta_mod_date = int(get_mod_date(media.markdown_metadata_filepath))
            except FileNotFoundError:
                pass
            else:
                data["meta_mod_date"] = meta_mod_date
                data["markdown_metadata"] = media.markdown_metadata

            cache[os.path.join(media.path, media.dst_filename)] = data

    cachePath = os.path.join(gallery.settings["destination"], ".metadata_cache")

    if len(cache) == 0:
        if os.path.exists(cachePath):
            os.remove(cachePath)
        return

    try:
        with open(cachePath, "wb") as cacheFile:
            pickle.dump(cache, cacheFile)
            logger.debug("Stored cache with %d entries", len(gallery.metadataCache))
    except Exception as e:
        logger.warn("Could not store cache: %s", e)
        os.remove(cachePath)


def register(settings):
    signals.gallery_build.connect(save_cache)
    signals.album_initialized.connect(load_metadata)



# Copyright (c) 2011-2023 - Simon Conseil

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import importlib
import logging
import os
import shutil
import sys
from fnmatch import fnmatch
from functools import lru_cache
from urllib.parse import quote

from markdown import Markdown
from markupsafe import Markup

from sigal.settings import Status

logger = logging.getLogger(__name__)
MD = None
VIDEO_MIMES = {".mp4": "video/mp4", ".webm": "video/webm", ".ogv": "video/ogg"}


class Devnull:
    """'Black hole' for output that should not be printed"""

    def write(self, *_):
        pass

    def flush(self, *_):
        pass


def copy(src, dst, symlink=False, rellink=False):
    """Copy or symlink the file."""
    func = os.symlink if symlink else shutil.copy2
    if symlink and os.path.lexists(dst):
        os.remove(dst)
    if rellink:  # relative symlink from dst
        func(os.path.relpath(src, os.path.dirname(dst)), dst)
    else:
        try:
            func(src, dst)
        except PermissionError:
            # this can happen if the file is not writable, so we try to remove
            # it first
            os.remove(dst)
            func(src, dst)


def check_or_create_dir(path):
    "Create the directory if it does not exist"

    if not os.path.isdir(path):
        os.makedirs(path)


@lru_cache(maxsize=1024)
def get_mod_date(path):
    """Get modification date for a path, caching result with LRU cache."""
    return os.path.getmtime(path)


def url_from_path(path):
    """Transform path to url, converting backslashes to slashes if needed."""

    if os.sep != "/":
        path = "/".join(path.split(os.sep))
    return quote(path)


def should_reprocess_album(path, name, force=False):
    if isinstance(force, bool):
        return force
    else:
        for f in force:
            if "*" in f or "?" in f:
                if fnmatch(path, f):
                    return True
            elif name == f:
                return True
    return False


def read_markdown(filename):
    """Reads markdown file, converts output and fetches title and meta-data for
    further processing.
    """
    global MD
    # Use utf-8-sig codec to remove BOM if it is present. This is only possible
    # this way prior to feeding the text to the markdown parser (which would
    # also default to pure utf-8)
    with open(filename, encoding="utf-8-sig") as f:
        text = f.read()

    if MD is None:
        MD = Markdown(
            extensions=[
                "markdown.extensions.extra",
                "markdown.extensions.meta",
                "markdown.extensions.tables",
            ],
            output_format="html5",
        )
    else:
        MD.reset()
        # When https://github.com/Python-Markdown/markdown/pull/672
        # will be available, this can be removed.
        MD.Meta = {}

    # Mark HTML with Markup to prevent jinja2 autoescaping
    output = {"description": Markup(MD.convert(text))}

    try:
        meta = MD.Meta.copy()
    except AttributeError:
        pass
    else:
        output["meta"] = meta
        try:
            output["title"] = MD.Meta["title"][0]
        except KeyError:
            pass

    return output


def is_valid_html5_video(ext):
    """Checks if ext is a supported HTML5 video."""
    return ext in VIDEO_MIMES.keys()


def get_mime(ext):
    """Returns mime type for extension."""
    return VIDEO_MIMES[ext]


def init_plugins(settings):
    """Load plugins and call register()."""

    logger = logging.getLogger(__name__)
    logger.debug("Plugin paths: %s", settings["plugin_paths"])

    for path in settings["plugin_paths"]:
        sys.path.insert(0, path)

    for plugin in settings["plugins"]:
        try:
            if isinstance(plugin, str):
                mod = importlib.import_module(plugin)
                mod.register(settings)
            else:
                plugin.register(settings)
            logger.debug("Registered plugin %s", plugin)
        except Exception as e:
            logger.error("Failed to load plugin %s: %r", plugin, e)

    for path in settings["plugin_paths"]:
        sys.path.remove(path)


class raise_if_debug:
    def __init__(self):
        self.value = None

    def __enter__(self, *args):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type:
            logger.info("Failed to process: %r", exc_value)
            if logger.getEffectiveLevel() == logging.DEBUG:
                # propagate the exception
                return False
            else:
                self.value = Status.FAILURE
        else:
            self.value = Status.SUCCESS

        # suppress the exception
        return True



# Copyright 2019 - Remi Ferrand

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

""" This plugin controls the generation of a ZIP archive for a gallery

If the ``zip_gallery`` setting is set, it contains the location of a zip
archive with all original images of the corresponding directory.

To ignore a ZIP gallery generation for a particular album, put
a ``.nozip_gallery`` file next to it in its parent folder. Only the existence
of this ``.nozip_gallery`` file is tested.  If no ``.nozip_gallery`` file is
present, then make a ZIP archive with all media files.

See :ref:`compatibility with the encrypt plugin <compatibility-with-encrypt>`.
"""

import logging
import os
import zipfile
from functools import cached_property
from os.path import isfile, join

from sigal import signals
from sigal.gallery import Album

logger = logging.getLogger(__name__)


def _should_generate_album_zip(album):
    """Checks whether a `.nozip_gallery` file exists in the album folder"""
    nozipgallerypath = os.path.join(album.src_path, ".nozip_gallery")
    return not os.path.isfile(nozipgallerypath)


def _generate_album_zip(album):
    """Make a ZIP archive with all media files and return its path.

    If the ``zip_gallery`` setting is set,it contains the location of a zip
    archive with all original images of the corresponding directory.
    """

    zip_gallery = album.settings["zip_gallery"]

    if zip_gallery and len(album) > 0:
        zip_gallery = zip_gallery.format(album=album)
        archive_path = join(album.dst_path, zip_gallery)
        if album.settings.get("zip_skip_if_exists", False) and isfile(archive_path):
            logger.debug("Archive %s already created, passing", archive_path)
            return zip_gallery

        archive = zipfile.ZipFile(archive_path, "w", allowZip64=True)
        attr = (
            "src_path" if album.settings["zip_media_format"] == "orig" else "dst_path"
        )

        for p in album:
            path = getattr(p, attr)
            try:
                archive.write(path, os.path.split(path)[1])
            except OSError as e:
                logger.warn("Failed to add %s to the ZIP: %s", p, e)

        archive.close()
        logger.debug("Created ZIP archive %s", archive_path)
        return zip_gallery

    return False


def generate_album_zip(album):
    """Checks for .nozip_gallery file in album folder.
    If this file exists, no ZIP archive is generated.
    If the file is absent, make a ZIP archive with all media files and return its path.

    If the ``zip_gallery`` setting is set,it contains the location of a zip
    archive with all original images of the corresponding directory.
    """

    # check if ZIP file generation as been disabled by .nozip_gallery file
    if not _should_generate_album_zip(album):
        logger.info(
            "Ignoring ZIP gallery generation for album '%s' because of present "
            ".nozip_gallery file",
            album.name,
        )
        return False

    return _generate_album_zip(album)


def nozip_gallery_file(album, settings=None):
    """Filesystem based switch to disable ZIP generation for an Album"""
    Album.zip = cached_property(generate_album_zip)
    Album.zip.__set_name__(Album, "zip")


def check_settings(gallery):
    if gallery.settings["zip_gallery"] and not isinstance(
        gallery.settings["zip_gallery"], str
    ):
        logger.error("'zip_gallery' should be set to a filename")
        gallery.settings["zip_gallery"] = False


def register(settings):
    signals.album_initialized.connect(nozip_gallery_file)
    signals.gallery_initialized.connect(check_settings)



"""
Plugin to protect gallery by encrypting image files using a password.

Options::

    encrypt_options = {
        'password': 'password',
        'ask_password': False,
        'gcm_tag': 'randomly_generated',
        'kdf_salt': 'randomly_generated',
        'kdf_iters': 10000,
    }

- ``password``: The password used to encrypt the images on gallery build,
  and decrypt them when viewers access the gallery. No default value. You must
  specify a password.
- ``ask_password``: Whether or not viewers are asked for the password to view
  the gallery. If set to ``False``, the password will be present in the HTML
  files so the images are decrypted automatically. Defaults to ``False``.
- ``gcm_tag``, ``kdf_salt``, ``kdf_iters``: Cryptographic parameters used when
  encrypting the files. ``gcm_tag``, ``kdf_salt`` are meant to be randomly
  generated, ``kdf_iters`` defaults to 10000. Do not specify them in the config
  file unless you have good reasons to do so.

Note: The plugin caches the cryptographic parameters (but not the password)
after the first build, so that incremental builds can share the same
credentials.  DO NOT CHANGE THE PASSWORD OR OTHER CRYPTOGRAPHIC PARAMETERS ONCE
A GALLERY IS BUILT, or there will be inconsistency in encrypted files and
viewers will not be able to see some of the images any more.

.. _compatibility-with-encrypt:

Compatibility with other plugins:

- ``zip_gallery``: if you enable both this plugin and the ``zip_gallery``
  plugin, the generated zip archives will contain encrypted images, which is
  generally meaningless since viewers cannot easily decrypt them outside
  a browser.

"""

from .encrypt import register  # noqa



"""Plugin which adjust the image after resizing.

Based on pilkit's Adjust_ processor.

.. _Adjust: \
   https://github.com/matthewwithanm/pilkit/blob/master/pilkit/processors/base.py#L19

Settings::

    adjust_options = {
        'color': 1.0,
        'brightness': 1.0,
        'contrast': 1.0,
        'sharpness': 1.0
    }

"""

import logging

from pilkit.processors import Adjust

from sigal import signals

logger = logging.getLogger(__name__)


def adjust(img, settings=None):
    logger.debug("Adjust image %r", img)
    return Adjust(**settings["adjust_options"]).process(img)


def register(settings):
    if settings.get("adjust_options"):
        signals.img_resized.connect(adjust)
    else:
        logger.warning("Adjust options are not set")



# Copyright (c) 2009-2023 - Simon Conseil
# Copyright (c) 2013      - Christophe-Marie Duquesne
# Copyright (c) 2017      - Mate Lakat
# Copyright (c) 2021      - Keith Feldman

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import logging
import os
from os.path import abspath, isabs, join, normpath
from pprint import pformat

_DEFAULT_CONFIG = {
    "albums_sort_attr": "name",
    "albums_sort_reverse": False,
    "autorotate_images": True,
    "colorbox_column_size": 3,
    "copy_exif_data": False,
    "datetime_format": "%c",
    "destination": "_build",
    "files_to_copy": (),
    "galleria_theme": "classic",
    "google_analytics": "",
    "google_tag_manager": "",
    "ignore_directories": [],
    "ignore_files": [],
    "img_extensions": [".jpg", ".jpeg", ".png", ".gif", ".tif", ".tiff", ".webp"],
    "img_processor": "ResizeToFit",
    "img_size": (640, 480),
    "img_format": None,
    "index_in_url": False,
    "jpg_options": {"quality": 85, "optimize": True, "progressive": True},
    "keep_orig": False,
    "html_language": "en",
    "leaflet_provider": "OpenStreetMap.Mapnik",
    "links": "",
    "locale": "",
    "make_thumbs": True,
    "max_img_pixels": None,
    "map_height": "500px",
    "medias_sort_attr": "filename",
    "medias_sort_reverse": False,
    "mp4_options": ["-crf", "23", "-strict", "-2"],
    "mp4_options_second_pass": None,
    "orig_dir": "original",
    "orig_link": False,
    "rel_link": False,
    "output_filename": "index.html",
    "piwik": {"tracker_url": "", "site_id": 0},
    "plugin_paths": [],
    "plugins": [],
    "site_logo": "",
    "show_map": False,
    "source": "",
    "theme": "colorbox",
    "thumb_dir": "thumbnails",
    "thumb_fit": True,
    "thumb_fit_centering": (0.5, 0.5),
    "thumb_prefix": "",
    "thumb_size": (200, 150),
    "thumb_suffix": "",
    "thumb_video_delay": 0,
    "thumb_video_black_retries": 0,
    "thumb_video_black_retry_offset": 1,
    "thumb_video_black_max_colors": 4,
    "title": "",
    "use_orig": False,
    "user_css": None,
    "video_converter": "ffmpeg",
    "video_extensions": [".3gp", ".avi", ".mkv", ".mov", ".mp4", ".ogv", ".webm"],
    "video_format": "webm",
    "video_always_convert": False,
    "video_size": (480, 360),
    "watermark": "",
    "webm_options": ["-crf", "10", "-b:v", "1.6M", "-qmin", "4", "-qmax", "63"],
    "webm_options_second_pass": None,
    "write_html": True,
    "zip_gallery": False,
    "zip_media_format": "resized",
}


class Status:
    SUCCESS = 0
    FAILURE = 1


def get_thumb(settings, filename):
    """Return the path to the thumb.

    examples:
    >>> default_settings = create_settings()
    >>> get_thumb(default_settings, "bar/foo.jpg")
    "bar/thumbnails/foo.jpg"
    >>> get_thumb(default_settings, "bar/foo.png")
    "bar/thumbnails/foo.png"

    for videos, it returns a jpg file:
    >>> get_thumb(default_settings, "bar/foo.webm")
    "bar/thumbnails/foo.jpg"
    """

    path, filen = os.path.split(filename)
    name, ext = os.path.splitext(filen)

    if ext.lower() in settings["video_extensions"]:
        ext = ".jpg"
    return join(
        path,
        settings["thumb_dir"],
        settings["thumb_prefix"] + name + settings["thumb_suffix"] + ext,
    )


def read_settings(filename=None):
    """Read settings from a config file in the source_dir root."""

    logger = logging.getLogger(__name__)
    logger.info("Reading settings ...")
    settings = _DEFAULT_CONFIG.copy()

    if filename:
        logger.debug("Settings file: %s", filename)
        settings_path = os.path.dirname(filename)
        tempdict = {}

        with open(filename) as f:
            code = compile(f.read(), filename, "exec")
            exec(code, tempdict)

        settings.update(
            (k, v) for k, v in tempdict.items() if k not in ["__builtins__"]
        )

        # Make the paths relative to the settings file
        paths = ["source", "destination", "watermark"]

        if os.path.isdir(join(settings_path, settings["theme"])) and os.path.isdir(
            join(settings_path, settings["theme"], "templates")
        ):
            paths.append("theme")

        for p in paths:
            path = settings[p]
            if path and not isabs(path):
                settings[p] = abspath(normpath(join(settings_path, path)))
                logger.debug("Rewrite %s : %s -> %s", p, path, settings[p])

    for key in ("img_size", "thumb_size", "video_size"):
        if settings[key]:
            w, h = settings[key]
            if h > w:
                settings[key] = (h, w)
                logger.warning(
                    "The %s setting should be specified with the largest value first.",
                    key,
                )

    if not settings["img_processor"]:
        logger.info("No Processor, images will not be resized")

    logger.debug("Settings:\n%s", pformat(settings, width=120))
    return settings


def create_settings(**kwargs):
    """Create a new default setting copy and initialize it with kwargs."""
    settings = _DEFAULT_CONFIG.copy()
    settings.update(kwargs)
    return settings



from blinker import signal

img_resized = signal("img_resized")

album_initialized = signal("album_initialized")
gallery_initialized = signal("gallery_initialized")
gallery_build = signal("gallery_build")
media_initialized = signal("media_initialized")
albums_sorted = signal("albums_sorted")
medias_sorted = signal("medias_sorted")
before_render = signal("before_render")
album_file = signal("album_file")
process_file = signal("process_file")



#!/usr/bin/env python3

# copyright (c) 2020 Bowen Ding

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import io
import os
from base64 import b64decode
from pathlib import Path
from typing import BinaryIO

from cryptography.exceptions import InvalidTag
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.ciphers.aead import AESGCM
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC

backend = default_backend()
MAGIC_STRING = b"_e_n_c_r_y_p_t_e_d_"


def kdf_gen_key(password: str, salt: str, iters: int) -> bytes:
    password = password.encode("utf-8")
    salt = salt.encode("utf-8")
    kdf = PBKDF2HMAC(
        algorithm=hashes.SHA1(), length=16, salt=salt, iterations=iters, backend=backend
    )
    key = kdf.derive(password)
    return key


def dispatchargs(decorated):
    def wrapper(args):
        if args.key is not None:
            key = b64decode(args.key.encode("utf-8"))
        elif args.password is not None:
            key = kdf_gen_key(args.password, args.kdf_salt, args.kdf_iters)
        else:
            raise ValueError("Neither password nor key is provided")
        tag = args.gcm_tag.encode("utf-8")
        outputBuffer = io.BytesIO()
        with Path(args.infile).open("rb") as in_:
            decorated(key, in_, outputBuffer, tag)
        with Path(args.outfile).open("wb") as out:
            out.write(outputBuffer.getbuffer())

    return wrapper


def encrypt(key: bytes, infile: BinaryIO, outfile: BinaryIO, tag: bytes):
    if len(key) != 128 / 8:
        raise ValueError("Unsupported key length: %d" % len(key))
    aesgcm = AESGCM(key)
    iv = os.urandom(12)
    plaintext = infile
    ciphertext = outfile
    rawbytes = plaintext.read()
    encrypted = aesgcm.encrypt(iv, rawbytes, tag)
    ciphertext.write(MAGIC_STRING)
    ciphertext.write(iv)
    ciphertext.write(encrypted)


def decrypt(key: bytes, infile: BinaryIO, outfile: BinaryIO, tag: bytes):
    if len(key) != 128 / 8:
        raise ValueError("Unsupported key length: %d" % len(key))
    aesgcm = AESGCM(key)
    ciphertext = infile
    plaintext = outfile
    magicstring = ciphertext.read(len(MAGIC_STRING))
    if magicstring != MAGIC_STRING:
        raise ValueError("Data is not encrypted")
    iv = ciphertext.read(12)
    rawbytes = ciphertext.read()
    try:
        decrypted = aesgcm.decrypt(iv, rawbytes, tag)
    except InvalidTag:
        raise ValueError("Incorrect tag, iv, or corrupted ciphertext")
    plaintext.write(decrypted)


if __name__ == "__main__":
    import argparse as ap

    parser = ap.ArgumentParser(description="Encrypt or decrypt using AES-128-GCM")
    parser.add_argument("-k", "--key", help="Base64-encoded key")
    parser.add_argument("-p", "--password", help="Password in plaintext")
    parser.add_argument(
        "--kdf-salt", help="PBKDF2 salt", default="saltysaltsweetysweet"
    )
    parser.add_argument(
        "--kdf-iters", type=int, help="PBKDF2 iterations", default=10000
    )
    parser.add_argument("--gcm-tag", help="AES-GCM tag", default="AuTheNTiCatIoNtAG")
    parser.add_argument("-i", "--infile", help="Input file")
    parser.add_argument("-o", "--outfile", help="Output file")
    subparsers = parser.add_subparsers(title="commands", dest="action")
    parser_enc = subparsers.add_parser("enc", help="Encrypt")
    parser_enc.set_defaults(execute=dispatchargs(encrypt))
    parser_dec = subparsers.add_parser("dec", help="Decrypt")
    parser_dec.set_defaults(execute=dispatchargs(decrypt))

    args = parser.parse_args()
    args.execute(args)



# Copyright (c) 2009-2023 - Simon Conseil
# Copyright (c)      2013 - Christophe-Marie Duquesne
# Copyright (c)      2018 - Edwin Steele

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import importlib
import logging
import os
import shutil
import stat
import sys
import types

from jinja2 import ChoiceLoader, Environment, FileSystemLoader, PrefixLoader
from jinja2.exceptions import TemplateNotFound

from . import signals
from .utils import url_from_path

THEMES_PATH = os.path.normpath(
    os.path.join(os.path.abspath(os.path.dirname(__file__)), "themes")
)


class AbstractWriter:
    template_file = None

    def __init__(self, settings, index_title=""):
        self.settings = settings
        self.output_dir = settings["destination"]
        self.theme = settings["theme"]
        self.index_title = index_title
        self.logger = logging.getLogger(__name__)

        # search the theme in sigal/theme if the given one does not exists
        if not os.path.exists(self.theme) or not os.path.exists(
            os.path.join(self.theme, "templates")
        ):
            self.theme = os.path.join(THEMES_PATH, self.theme)
            if not os.path.exists(self.theme):
                raise Exception("Impossible to find the theme %s" % self.theme)

        self.logger.info("Theme  : %s", self.theme)
        theme_relpath = os.path.join(self.theme, "templates")
        default_loader = FileSystemLoader(
            os.path.join(THEMES_PATH, "default", "templates")
        )

        # setup jinja env
        env_options = {"trim_blocks": True, "autoescape": True, "lstrip_blocks": True}

        loaders = [
            FileSystemLoader(theme_relpath),
            default_loader,  # implicit inheritance
            PrefixLoader({"!default": default_loader}),  # explicit one
        ]
        env = Environment(loader=ChoiceLoader(loaders), **env_options)

        # handle optional filters.py
        filters_py = os.path.join(self.theme, "filters.py")
        if os.path.exists(filters_py):
            self.logger.info("Loading filters file: %s", filters_py)
            module_spec = importlib.util.spec_from_file_location("filters", filters_py)
            mod = importlib.util.module_from_spec(module_spec)
            sys.modules["filters"] = mod
            module_spec.loader.exec_module(mod)
            for name in dir(mod):
                if isinstance(getattr(mod, name), types.FunctionType):
                    env.filters[name] = getattr(mod, name)

        try:
            self.template = env.get_template(self.template_file)
        except TemplateNotFound:
            self.logger.error(
                "The template %s was not found in template folder %s.",
                self.template_file,
                theme_relpath,
            )
            sys.exit(1)

        # Copy the theme files in the output dir
        self.theme_path = os.path.join(self.output_dir, "static")
        if os.path.isdir(self.theme_path):
            shutil.rmtree(self.theme_path)

        for static_path in (
            os.path.join(THEMES_PATH, "default", "static"),
            os.path.join(self.theme, "static"),
        ):
            shutil.copytree(static_path, self.theme_path, dirs_exist_ok=True)

            # Ensure that the theme dir is writeable
            for root, _, files in os.walk(self.theme_path):
                st = os.stat(root)
                os.chmod(root, st.st_mode | stat.S_IWUSR)
                for name in files:
                    path = os.path.join(root, name)
                    st = os.stat(path)
                    os.chmod(path, st.st_mode | stat.S_IWUSR)

        if self.settings["user_css"]:
            if not os.path.exists(self.settings["user_css"]):
                self.logger.error(
                    "CSS file %s could not be found", self.settings["user_css"]
                )
            else:
                shutil.copy(self.settings["user_css"], self.theme_path)

    def generate_context(self, album):
        """Generate the context dict for the given path."""

        from . import __url__ as sigal_link

        self.logger.info("Output album : %r", album)
        ctx = {
            "album": album,
            "index_title": self.index_title,
            "settings": self.settings,
            "sigal_link": sigal_link,
            "theme": {
                "name": os.path.basename(self.theme),
                "url": url_from_path(os.path.relpath(self.theme_path, album.dst_path)),
            },
        }
        if self.settings["user_css"]:
            ctx["user_css"] = os.path.basename(self.settings["user_css"])
        return ctx

    def write(self, album):
        """Generate the HTML page and save it."""
        context = self.generate_context(album)
        signals.before_render.send(context)
        page = self.template.render(**context)
        output_file = os.path.join(album.dst_path, album.output_file)

        with open(output_file, "w", encoding="utf-8") as f:
            f.write(page)


class AlbumListPageWriter(AbstractWriter):
    """Generate an html page for a directory of albums"""

    template_file = "album_list.html"


class AlbumPageWriter(AbstractWriter):
    """Generate html pages for a directory of images."""

    template_file = "album.html"



# Copyright (c) 2009-2023 - Simon Conseil

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import locale
import logging
import os
import pathlib
import socketserver
import sys
import time
from http import server

import click
from click import argument, option

from .gallery import Gallery
from .log import init_logging
from .settings import read_settings
from .utils import copy, init_plugins

try:
    from .version import __version__
except ImportError:
    # package is not installed
    __version__ = None

_DEFAULT_CONFIG_FILE = "sigal.conf.py"


@click.group()
@click.version_option(version=__version__)
def main():
    """Sigal - Simple Static Gallery Generator.

    Sigal is yet another python script to prepare a static gallery of images:
    resize images, create thumbnails with some options, generate html pages.

    """


@main.command()
@argument("path", default=_DEFAULT_CONFIG_FILE)
def init(path):
    """Copy a sample config file in the current directory (default to
    'sigal.conf.py'), or use the provided 'path'."""

    path = pathlib.Path(path)
    if path.exists():
        print("Found an existing config file, will abort to keep it safe.")
        sys.exit(1)

    conf = pathlib.Path(__file__).parent / "templates" / "sigal.conf.py"
    path.write_text(conf.read_text())
    print(f"Sample config file created: {path}")


@main.command()
@argument("source", required=False)
@argument("destination", required=False)
@option("-f", "--force", is_flag=True, help="Force the reprocessing of existing images")
@option(
    "-a",
    "--force-album",
    multiple=True,
    help=(
        "Force reprocessing of any album that matches the given pattern. "
        "Patterns containing no wildcards will be matched against only "
        "the album name. (-a 'My Pictures/* Pics' -a 'Festival')"
    ),
)
@option("-v", "--verbose", is_flag=True, help="Show all messages")
@option(
    "-d",
    "--debug",
    is_flag=True,
    help=(
        "Show all messages, including debug messages. Also raise "
        "exception if an error happen when processing files."
    ),
)
@option("-q", "--quiet", is_flag=True, help="Show only error messages")
@option(
    "-c",
    "--config",
    default=_DEFAULT_CONFIG_FILE,
    show_default=True,
    help="Configuration file",
)
@option(
    "-t",
    "--theme",
    help=(
        "Specify a theme directory, or a theme name for the themes included with Sigal"
    ),
)
@option("--title", help="Title of the gallery (overrides the title setting.")
@option("-n", "--ncpu", help="Number of cpu to use (default: all)")
def build(
    source,
    destination,
    debug,
    verbose,
    quiet,
    force,
    force_album,
    config,
    theme,
    title,
    ncpu,
):
    """Run sigal to process a directory.

    If provided, 'source', 'destination' and 'theme' will override the
    corresponding values from the settings file.

    """
    if sum([debug, verbose, quiet]) > 1:
        sys.exit("Only one option of debug, verbose and quiet should be used")

    show_progress = False
    if debug:
        level = logging.DEBUG
    elif verbose:
        level = logging.INFO
    elif quiet:
        level = logging.ERROR
    else:
        level = logging.WARNING
        show_progress = True

    init_logging("sigal", level=level)
    logger = logging.getLogger(__name__)

    if not os.path.isfile(config):
        logger.error("Settings file not found: %s", config)
        sys.exit(1)

    start_time = time.time()
    settings = read_settings(config)

    for key in ("source", "destination", "theme"):
        arg = locals()[key]
        if arg is not None:
            settings[key] = os.path.abspath(arg)
        logger.info("%12s : %s", key.capitalize(), settings[key])

    if not settings["source"] or not os.path.isdir(settings["source"]):
        logger.error("Input directory not found: %s", settings["source"])
        sys.exit(1)

    # on windows os.path.relpath raises a ValueError if the two paths are on
    # different drives, in that case we just ignore the exception as the two
    # paths are anyway not relative
    relative_check = True
    try:
        relative_check = os.path.relpath(
            settings["destination"], settings["source"]
        ).startswith("..")
    except ValueError:
        pass

    if not relative_check:
        logger.error("Output directory should be outside of the input directory.")
        sys.exit(1)

    if title:
        settings["title"] = title

    locale.setlocale(locale.LC_ALL, settings["locale"])
    init_plugins(settings)

    gal = Gallery(settings, ncpu=ncpu, show_progress=show_progress)
    gal.build(force=force_album if len(force_album) else force)

    # copy extra files
    for src, dst in settings["files_to_copy"]:
        src = os.path.join(settings["source"], src)
        dst = os.path.join(settings["destination"], dst)
        logger.debug("Copy %s to %s", src, dst)
        copy(src, dst, symlink=settings["orig_link"], rellink=settings["rel_link"])

    stats = gal.stats

    def format_stats(_type):
        opt = [
            "{} {}".format(stats[_type + "_" + subtype], subtype)
            for subtype in ("skipped", "failed")
            if stats[_type + "_" + subtype] > 0
        ]
        opt = " ({})".format(", ".join(opt)) if opt else ""
        return f"{stats[_type]} {_type}s{opt}"

    if not quiet and len(stats) > 0:
        stats_str = ""
        types = sorted({t.rsplit("_", 1)[0] for t in stats})
        for t in types[:-1]:
            stats_str += f"{format_stats(t)} and "
        stats_str += f"{format_stats(types[-1])}"
        end_time = time.time() - start_time
        print(f"Done, processed {stats_str} in {end_time:.2f} seconds.")


@main.command()
@argument("destination", default="_build")
@option("-p", "--port", help="Port to use", default=8000)
@option(
    "-c",
    "--config",
    default=_DEFAULT_CONFIG_FILE,
    show_default=True,
    help="Configuration file",
)
def serve(destination, port, config):
    """Run a simple web server."""
    if os.path.exists(destination):
        pass
    elif os.path.exists(config):
        settings = read_settings(config)
        destination = settings.get("destination")
        if not os.path.exists(destination):
            sys.stderr.write(
                f"The '{destination}' directory doesn't exist, maybe try building"
                " first?\n"
            )
            sys.exit(1)
    else:
        sys.stderr.write(
            f"The {destination} directory doesn't exist "
            f"and the config file ({config}) could not be read.\n"
        )
        sys.exit(2)

    print(f"DESTINATION : {destination}")
    os.chdir(destination)
    Handler = server.SimpleHTTPRequestHandler
    httpd = socketserver.TCPServer(("", port), Handler, False)
    print(f" * Running on http://127.0.0.1:{port}/")

    try:
        httpd.allow_reuse_address = True
        httpd.server_bind()
        httpd.server_activate()
        httpd.serve_forever()
    except KeyboardInterrupt:
        print("\nAll done!")


@main.command()
@argument("target")
@argument("keys", nargs=-1)
@option(
    "-o", "--overwrite", default=False, is_flag=True, help="Overwrite existing .md file"
)
def set_meta(target, keys, overwrite=False):
    """Write metadata keys to .md file.

    TARGET can be a media file or an album directory. KEYS are key/value pairs.

    Ex, to set the title of test.jpg to "My test image":

    sigal set_meta test.jpg title "My test image"
    """

    if not os.path.exists(target):
        sys.stderr.write(f"The target {target} does not exist.\n")
        sys.exit(1)
    if len(keys) < 2 or len(keys) % 2 > 0:
        sys.stderr.write("Need an even number of arguments.\n")
        sys.exit(1)

    if os.path.isdir(target):
        descfile = os.path.join(target, "index.md")
    else:
        descfile = os.path.splitext(target)[0] + ".md"
    if os.path.exists(descfile) and not overwrite:
        sys.stderr.write(
            f"Description file '{descfile}' already exists. "
            "Use --overwrite to overwrite it.\n"
        )
        sys.exit(2)

    with open(descfile, "w") as fp:
        for i in range(len(keys) // 2):
            k, v = keys[i * 2 : (i + 1) * 2]
            fp.write(f"{k.capitalize()}: {v}\n")
    print(f"{len(keys) // 2} metadata key(s) written to {descfile}")


if __name__ == "__main__":
    main()



"""Plugin which add a copyright to the image.

Settings:

- ``copyright``: the copyright text.
- ``copyright_text_font``: the copyright text font - either system/user
  font-name or absolute path to font.ttf file.  If no font is specified, or
  specified font is not found, the default font is used.
- ``copyright_text_font_size``: the copyright text font-size. If no font is
  specified, this setting is ignored.
- ``copyright_text_color``: the copyright text color in a tuple (R, G, B)
  with decimal RGB code, e.g. ``(255, 255, 255)`` is white.
- ``copyright_text_position``: the copyright text position in 2 tuple (left,
  top).  By default text would be positioned at bottom-left corner.

"""

import logging

from PIL import ImageDraw, ImageFont

from sigal import signals

logger = logging.getLogger(__name__)


def add_copyright(img, settings=None):
    logger.debug("Adding copyright to %r", img)
    draw = ImageDraw.Draw(img)
    text = settings["copyright"]
    font = settings.get("copyright_text_font", None)
    font_size = settings.get("copyright_text_font_size", 10)
    assert font_size >= 0
    color = settings.get("copyright_text_color", (0, 0, 0))
    bottom_margin = 3  # bottom margin for text
    text_height = bottom_margin + 12  # default text height (of 15)
    if font:
        try:
            font = ImageFont.truetype(font, font_size)
            text_height = font.getsize(text)[1] + bottom_margin
        except Exception:  # load default font in case of any exception
            logger.debug("Exception: Couldn't locate font %s, using default font", font)
            font = ImageFont.load_default()
    else:
        font = ImageFont.load_default()
    left, top = settings.get("copyright_text_position", (5, img.size[1] - text_height))
    draw.text((left, top), text, fill=color, font=font)
    return img


def register(settings):
    if settings.get("copyright"):
        signals.img_resized.connect(add_copyright)
    else:
        logger.warning("Copyright text is not set")



"""Plugin to index non-media files.

This plugin will copy the files into the build tree and generate generic
thumbnails for the files.  In-browser previews will likely fail, and
it is up to the theme to provide correct support for downloads.

Settings available as dictionary in ``nonmedia_files_options``:

- ``ext_as_thumb``: Enable simple thumbnail showing ext. Default to ``True``
- ``ignore_ext``: List of file extensions to ignore. Default to ``[".md"]``
- ``thumb_bg_color``: Background color for thumbnail. Accepts (r, g, b) tuple.
  Default to white ``(255, 255, 255)``.
- ``thumb_font``: Name or path to font file.
  Default to ``None`` to use the PIL built-in font.
- ``thumb_font_color``: Font color for thumbnail. Accepts (r, g, b) tuple.
  Default to black ``(0, 0, 0)``.
- ``thumb_font_size``: Font size for thumbnail. Must select a font to apply.
  Default to ``40``.

.. note:: Thumbnails are generated using the file extension text on a
    background color. It is highly recommended to select a font
    (such as ``"FreeMono.ttf"``), since the default PIL font does not
    respect font size.
"""

import logging
import os

from PIL import Image as PILImage
from PIL import ImageDraw, ImageFont
from pilkit.utils import save_image

from sigal import signals, utils
from sigal.gallery import Media

logger = logging.getLogger(__name__)


DEFAULT_CONFIG = {
    "ext_as_thumb": True,
    "ignore_ext": [".md"],
    "thumb_bg_color": (255, 255, 255),
    "thumb_font": None,
    "thumb_font_color": (0, 0, 0),
    "thumb_font_size": 40,
}


COMMON_MIME_TYPES = {
    ".azw": "application/vnd.amazon.ebook",
    ".csv": "text/csv",
    ".epub": "application/epub+zip",
    ".pdf": "application/pdf",
    ".svg": "image/svg+xml",
    ".txt": "text/plain",
    ".zip": "application/zip",
}


class NonMedia(Media):
    """Gather all informations on a non-media file."""

    type = "nonmedia"

    def __init__(self, filename, path, settings):
        super().__init__(filename, path, settings)
        self.thumb_name = os.path.splitext(self.thumb_name)[0] + ".jpg"
        self.date = self._get_file_date()
        self.mime = COMMON_MIME_TYPES.get(self.src_ext, "application/octet-stream")
        logger.debug("mime type %s", self.mime)

    @property
    def thumbnail(self):
        """Path to the thumbnail image (relative to the album directory)."""
        if not os.path.isfile(self.thumb_path):
            process_thumb(self)
        return super().thumbnail


def generate_thumbnail(
    text,
    outname,
    box,
    bg_color=DEFAULT_CONFIG["thumb_bg_color"],
    font=DEFAULT_CONFIG["thumb_font"],
    font_color=DEFAULT_CONFIG["thumb_font_color"],
    font_size=DEFAULT_CONFIG["thumb_font_size"],
    options=None,
):
    """Create a thumbnail image."""

    kwargs = {}
    if font:
        kwargs["font"] = ImageFont.truetype(font, font_size)
    if font_color:
        kwargs["fill"] = font_color

    img = PILImage.new("RGB", box, bg_color)

    anchor = (box[0] // 2, box[1] // 2)
    d = ImageDraw.Draw(img)
    logger.info(f"kwargs: {kwargs}")
    d.text(anchor, text, anchor="mm", **kwargs)

    outformat = "JPEG"
    logger.info("Save thumnail image: %s (%s)", outname, outformat)
    save_image(img, outname, outformat, options=options, autoconvert=True)


def process_thumb(media):
    settings = media.settings
    plugin_settings = settings.get("nonmedia_files_options", {})
    utils.copy(media.src_path, media.dst_path, symlink=settings["orig_link"])

    if plugin_settings.get("ext_as_thumb", DEFAULT_CONFIG["ext_as_thumb"]):
        logger.info("plugin_settings: %r", plugin_settings)
        kwargs = {}
        for key in ("bg_color", "font", "font_color", "font_size"):
            if f"thumb_{key}" in plugin_settings:
                kwargs[key] = plugin_settings[f"thumb_{key}"]
        generate_thumbnail(
            media.src_ext[1:].upper(),
            media.thumb_path,
            settings["thumb_size"],
            options=settings["jpg_options"],
            **kwargs,
        )


def process_nonmedia(media):
    """Process a non-media file: copy and create thumbnail."""
    logger.info("Processing non-media file: %s", media.dst_filename)

    with utils.raise_if_debug() as status:
        process_thumb(media)

    return status.value


def album_file(album, filename, media=None):
    if not media:
        ext = os.path.splitext(filename)[1]
        ext_ignore = album.settings.get("nonmedia_files_options", {}).get(
            "ignore_ext", DEFAULT_CONFIG["ignore_ext"]
        )
        if ext in ext_ignore:
            logger.info("Ignoring non-media file: %s", filename)
        else:
            logger.info("Registering non-media file: %s", filename)
            return NonMedia(filename, album.path, album.settings)


def process_file(media, processor=None):
    if media.type == "nonmedia":
        return process_nonmedia


def register(settings):
    signals.album_file.connect(album_file)
    signals.process_file.connect(process_file)



# Copyright (c) 2009-2023 - Simon Conseil
# Copyright (c) 2015 - François D.
# Copyright (c) 2018 - Edwin Steele

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

# Additional copyright notice:
#
# Several lines of code concerning extraction of GPS data from EXIF tags where
# taken from a GitHub Gist by Eran Sandler at
#
#   https://gist.github.com/erans/983821
#
# and partially modified. The code in question is licensed under MIT license.

import logging
import os
import sys
import warnings
from copy import deepcopy
from datetime import datetime

import pilkit.processors
from PIL import Image as PILImage
from PIL import ImageFile, ImageOps, IptcImagePlugin
from PIL.ExifTags import GPSTAGS, TAGS
from PIL.TiffImagePlugin import IFDRational
from pilkit.processors import Transpose
from pilkit.utils import save_image

from . import signals, utils

# Force loading of truncated files
ImageFile.LOAD_TRUNCATED_IMAGES = True


def _has_exif_tags(img):
    return hasattr(img, "info") and "exif" in img.info


def _read_image(file_path):
    # The image is already opened
    if isinstance(file_path, PILImage.Image):
        return file_path

    logger = logging.getLogger(__name__)

    with warnings.catch_warnings(record=True) as caught_warnings:
        im = PILImage.open(file_path)

    for w in caught_warnings:
        logger.warning(
            f"PILImage reported a warning for file {file_path}\n"
            f"{w.category}: {w.message}"
        )
    return im


def generate_image(source, outname, settings, options=None):
    """Image processor, rotate and resize the image.

    :param source: path to an image
    :param outname: output filename
    :param settings: settings dict
    :param options: dict with PIL options (quality, optimize, progressive)

    """

    logger = logging.getLogger(__name__)

    if settings["use_orig"] or source.endswith(".gif"):
        utils.copy(source, outname, symlink=settings["orig_link"])
        return

    img = _read_image(source)
    original_format = img.format

    if settings["copy_exif_data"] and settings["autorotate_images"]:
        logger.warning(
            "The 'autorotate_images' and 'copy_exif_data' settings "
            "are not compatible because Sigal can't save the "
            "modified Orientation tag."
        )

    # Preserve EXIF data
    if settings["copy_exif_data"] and _has_exif_tags(img):
        if options is not None:
            options = deepcopy(options)
        else:
            options = {}
        options["exif"] = img.info["exif"]

    # Rotate the img, and catch IOError when PIL fails to read EXIF
    if settings["autorotate_images"]:
        try:
            img = Transpose().process(img)
        except (OSError, IndexError):
            pass

    # Resize the image
    if settings["img_processor"]:
        try:
            logger.debug("Processor: %s", settings["img_processor"])
            processor_cls = getattr(pilkit.processors, settings["img_processor"])
        except AttributeError:
            logger.error("Wrong processor name: %s", settings["img_processor"])
            sys.exit()

        width, height = settings["img_size"]

        if img.size[0] < img.size[1]:
            # swap target size if image is in portrait mode
            height, width = width, height

        processor = processor_cls(width, height, upscale=False)
        img = processor.process(img)

    # signal.send() does not work here as plugins can modify the image, so we
    # iterate other the receivers to call them with the image.
    for receiver in signals.img_resized.receivers_for(img):
        img = receiver(img, settings=settings)

    # first, use hard-coded output format, or PIL format, or original image
    # format, or fall back to JPEG
    outformat = settings.get("img_format") or img.format or original_format or "JPEG"

    logger.debug("Save resized image to %s (%s)", outname, outformat)
    save_image(img, outname, outformat, options=options, autoconvert=True)


def generate_thumbnail(
    source, outname, box, fit=True, options=None, thumb_fit_centering=(0.5, 0.5)
):
    """Create a thumbnail image."""

    logger = logging.getLogger(__name__)
    img = _read_image(source)
    img = Transpose().process(img)
    original_format = img.format

    try:
        method = PILImage.Resampling.LANCZOS
    except AttributeError:
        # Deprecated since version 9.1.0
        method = PILImage.ANTIALIAS

    if fit:
        img = ImageOps.fit(img, box, method, centering=thumb_fit_centering)
    else:
        img.thumbnail(box, method)

    outformat = img.format or original_format or "JPEG"
    logger.debug("Save thumnail image: %s (%s)", outname, outformat)
    save_image(img, outname, outformat, options=options, autoconvert=True)


def process_image(media):
    """Process one image: resize, create thumbnail."""

    logger = logging.getLogger(__name__)
    logger.info("Processing %s", media.src_path)

    if media.src_ext in (".jpg", ".jpeg", ".JPG", ".JPEG"):
        options = media.settings["jpg_options"]
    elif media.src_ext == ".png":
        options = {"optimize": True}
    else:
        options = {}

    with utils.raise_if_debug() as status:
        generate_image(media.src_path, media.dst_path, media.settings, options=options)

        if media.settings["make_thumbs"]:
            generate_thumbnail(
                media.dst_path,
                media.thumb_path,
                media.settings["thumb_size"],
                fit=media.settings["thumb_fit"],
                options=options,
                thumb_fit_centering=media.settings["thumb_fit_centering"],
            )

    return status.value


def get_size(file_path):
    """Return image size (width and height)."""
    try:
        im = _read_image(file_path)
    except (OSError, IndexError, TypeError, AttributeError) as e:
        logger = logging.getLogger(__name__)
        logger.error("Could not read size of %s due to %r", file_path, e)
    else:
        width, height = im.size
        return {"width": width, "height": height}


def get_exif_data(filename):
    """Return a dict with the raw EXIF data."""

    logger = logging.getLogger(__name__)

    img = _read_image(filename)

    try:
        with warnings.catch_warnings(record=True) as caught_warnings:
            exif = img._getexif() or {}
    except ZeroDivisionError:
        logger.warning("Failed to read EXIF data.")
        return None

    for w in caught_warnings:
        fname = filename.filename if isinstance(filename, PILImage.Image) else filename
        logger.warning(
            f"PILImage reported a warning for file {fname}\n{w.category}: {w.message}"
        )

    data = {TAGS.get(tag, tag): value for tag, value in exif.items()}

    if "GPSInfo" in data:
        try:
            data["GPSInfo"] = {
                GPSTAGS.get(tag, tag): value for tag, value in data["GPSInfo"].items()
            }
        except AttributeError:
            logger.info("Failed to get GPS Info")
            del data["GPSInfo"]
    return data


def get_iptc_data(filename):
    """Return a dict with the raw IPTC data."""

    logger = logging.getLogger(__name__)

    iptc_data = {}
    raw_iptc = {}

    # PILs IptcImagePlugin issues a SyntaxError in certain circumstances
    # with malformed metadata, see PIL/IptcImagePlugin.py", line 71.
    # ( https://github.com/python-pillow/Pillow/blob/
    # 9dd0348be2751beb2c617e32ff9985aa2f92ae5f/src/PIL/IptcImagePlugin.py#L71 )
    try:
        img = _read_image(filename)
        raw_iptc = IptcImagePlugin.getiptcinfo(img)
    except SyntaxError:
        logger.info("IPTC Error in %s", filename)

    # IPTC fields are catalogued in:
    # https://www.iptc.org/std/photometadata/specification/IPTC-PhotoMetadata
    # 2:05 is the IPTC title property
    if raw_iptc and (2, 5) in raw_iptc:
        iptc_data["title"] = raw_iptc[(2, 5)].decode("utf-8", errors="replace")

    # 2:120 is the IPTC description property
    if raw_iptc and (2, 120) in raw_iptc:
        iptc_data["description"] = raw_iptc[(2, 120)].decode("utf-8", errors="replace")

    # 2:105 is the IPTC headline property
    if raw_iptc and (2, 105) in raw_iptc:
        iptc_data["headline"] = raw_iptc[(2, 105)].decode("utf-8", errors="replace")

    return iptc_data


def get_image_metadata(filename):
    logger = logging.getLogger(__name__)
    exif, iptc, size = {}, {}, {}
    img = None

    try:
        img = _read_image(filename)
    except Exception as e:
        logger.error("Could not open image %s metadata: %s", filename, e)
    else:
        try:
            if os.path.splitext(filename)[1].lower() in (".jpg", ".jpeg"):
                exif = get_exif_data(img)
        except Exception as e:
            logger.warning("Could not read EXIF data from %s: %s", filename, e)

        try:
            iptc = get_iptc_data(img)
        except Exception as e:
            logger.warning("Could not read IPTC data from %s: %s", filename, e)

        try:
            size = get_size(img)
        except Exception as e:
            logger.warning("Could not read size from %s: %s", filename, e)
    finally:
        if img is not None:
            img.close()

    return {"exif": exif, "iptc": iptc, "size": size}


def dms_to_degrees(v):
    """Convert degree/minute/second to decimal degrees."""

    d = float(v[0])
    m = float(v[1])
    s = float(v[2])
    return d + (m / 60.0) + (s / 3600.0)


def get_exif_tags(data, datetime_format="%c"):
    """Make a simplified version with common tags from raw EXIF data."""

    logger = logging.getLogger(__name__)
    simple = {}

    for tag in ("Model", "Make", "LensModel"):
        if tag in data:
            val = data[tag][0] if isinstance(data[tag], tuple) else data[tag]
            simple[tag] = str(val).strip()

    if "FNumber" in data:
        fnumber = data["FNumber"]
        try:
            simple["fstop"] = float(fnumber)
        except Exception:
            logger.debug("Skipped invalid FNumber: %r", fnumber, exc_info=True)

    if "FocalLength" in data:
        focal = data["FocalLength"]
        try:
            simple["focal"] = round(float(focal))
        except Exception:
            logger.debug("Skipped invalid FocalLength: %r", focal, exc_info=True)

    if "ExposureTime" in data:
        exptime = data["ExposureTime"]
        if isinstance(exptime, IFDRational):
            simple["exposure"] = f"{exptime.numerator}/{exptime.denominator}"
        elif isinstance(exptime, int):
            simple["exposure"] = str(exptime)
        else:
            logger.info("Unknown format for ExposureTime: %r", exptime)

    if data.get("ISOSpeedRatings"):
        simple["iso"] = data["ISOSpeedRatings"]

    if "DateTimeOriginal" in data:
        # Remove null bytes at the end if necessary
        date = data["DateTimeOriginal"].rsplit("\x00")[0]

        try:
            simple["dateobj"] = datetime.strptime(date, "%Y:%m:%d %H:%M:%S")
            simple["datetime"] = simple["dateobj"].strftime(datetime_format)
        except (ValueError, TypeError) as e:
            logger.info("Could not parse DateTimeOriginal: %s", e)

    if "GPSInfo" in data:
        info = data["GPSInfo"]
        lat_info = info.get("GPSLatitude")
        lon_info = info.get("GPSLongitude")
        lat_ref_info = info.get("GPSLatitudeRef")
        lon_ref_info = info.get("GPSLongitudeRef")

        if lat_info and lon_info and lat_ref_info and lon_ref_info:
            try:
                lat = dms_to_degrees(lat_info)
                lon = dms_to_degrees(lon_info)
            except (ZeroDivisionError, ValueError, TypeError):
                logger.info("Failed to read GPS info")
            else:
                simple["gps"] = {
                    "lat": -lat if lat_ref_info != "N" else lat,
                    "lon": -lon if lon_ref_info != "E" else lon,
                }

    return simple



# Copyright (c) 2009-2023 - Simon Conseil
# Copyright (c) 2013      - Christophe-Marie Duquesne
# Copyright (c) 2014      - Jonas Kaufmann
# Copyright (c) 2015      - François D.
# Copyright (c) 2017      - Mate Lakat
# Copyright (c) 2018      - Edwin Steele
# Copyright (c) 2021      - Tim AtLee

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import fnmatch
import io
import logging
import multiprocessing
import os
import pickle
import random
import sys
from collections import defaultdict
from datetime import datetime
from functools import cached_property
from itertools import cycle
from os.path import isfile, join, splitext
from shutil import get_terminal_size
from urllib.parse import quote as url_quote

from click import progressbar
from natsort import natsort_keygen, ns
from PIL import Image as PILImage

from . import image, signals, video
from .image import get_exif_tags, get_image_metadata, get_size, process_image
from .settings import Status, get_thumb
from .utils import (
    Devnull,
    check_or_create_dir,
    copy,
    get_mime,
    get_mod_date,
    is_valid_html5_video,
    read_markdown,
    should_reprocess_album,
    url_from_path,
)
from .video import process_video
from .writer import AlbumListPageWriter, AlbumPageWriter


class Media:
    """Base Class for media files.

    Attributes:

    :var Media.type: ``"image"`` or ``"video"``.
    :var Media.dst_filename: Filename of the resized image.
    :var Media.thumbnail: Location of the corresponding thumbnail image.
    :var Media.big: If not None, location of the unmodified image.
    :var Media.big_url: If not None, url of the unmodified image.

    """

    type = ""
    """Type of media, e.g. ``"image"`` or ``"video"``."""

    def __init__(self, filename, path, settings):
        self.path = path
        self.settings = settings

        self.basename = os.path.splitext(filename)[0]

        self.dst_filename = filename
        """Filename of the resized image."""

        self.src_filename = filename
        """Filename of the input image."""

        self.src_ext = os.path.splitext(filename)[1].lower()
        """Input extension."""

        self.src_path = join(settings["source"], path, self.src_filename)

        self.thumb_name = get_thumb(self.settings, self.dst_filename)

        self.logger = logging.getLogger(__name__)

        signals.media_initialized.send(self)

    def __repr__(self):
        return f"<{self.__class__.__name__}>({str(self)!r})"

    def __str__(self):
        return join(self.path, self.src_filename)

    def __getstate__(self):
        state = self.__dict__.copy()
        # remove un-pickable objects
        state["logger"] = None
        return state

    def __setstate__(self, state):
        for slot, value in state.items():
            setattr(self, slot, value)
        self.logger = logging.getLogger(__name__)

    @property
    def dst_path(self):
        return join(self.settings["destination"], self.path, self.dst_filename)

    @property
    def thumb_path(self):
        return join(self.settings["destination"], self.path, self.thumb_name)

    @property
    def url(self):
        """URL of the media."""
        return url_from_path(self.dst_filename)

    @property
    def big(self):
        """Path to the original image, if ``keep_orig`` is set (relative to the
        album directory). Copy the file if needed.
        """
        if self.settings["keep_orig"]:
            s = self.settings
            if s["use_orig"]:
                # The image *is* the original, just use it
                return self.src_filename
            orig_path = join(s["destination"], self.path, s["orig_dir"])
            check_or_create_dir(orig_path)
            big_path = join(orig_path, self.src_filename)
            if not isfile(big_path):
                copy(
                    self.src_path,
                    big_path,
                    symlink=s["orig_link"],
                    rellink=self.settings["rel_link"],
                )
            return join(s["orig_dir"], self.src_filename)

    @property
    def big_url(self):
        """URL of the original media."""
        if self.big is not None:
            return url_from_path(self.big)

    @property
    def thumbnail(self):
        """Path to the thumbnail image (relative to the album directory)."""

        if not isfile(self.thumb_path):
            self.logger.debug("Generating thumbnail for %r", self)
            path = self.dst_path if os.path.exists(self.dst_path) else self.src_path
            try:
                # if thumbnail is missing (if settings['make_thumbs'] is False)
                s = self.settings
                if self.type == "image":
                    image.generate_thumbnail(
                        path, self.thumb_path, s["thumb_size"], fit=s["thumb_fit"]
                    )
                elif self.type == "video":
                    video.generate_thumbnail(
                        path,
                        self.thumb_path,
                        s["thumb_size"],
                        s["thumb_video_delay"],
                        fit=s["thumb_fit"],
                        converter=s["video_converter"],
                        black_retries=s["thumb_video_black_retries"],
                        black_offset=s["thumb_video_black_retry_offset"],
                        black_max_colors=s["thumb_video_black_max_colors"],
                    )
            except Exception as e:
                self.logger.error("Failed to generate thumbnail: %s", e)
                return
        return url_from_path(self.thumb_name)

    @cached_property
    def description(self):
        """Description extracted from the Markdown <imagename>.md file."""
        return self.markdown_metadata.get("description", "")

    @cached_property
    def title(self):
        """Title extracted from the metadata, or defaults to the filename."""
        title = self.markdown_metadata.get("title", "")
        return title if title else self.basename

    @cached_property
    def meta(self):
        """Other metadata extracted from the Markdown <imagename>.md file."""
        return self.markdown_metadata.get("meta", {})

    @cached_property
    def markdown_metadata(self):
        """Get metadata from filename.md: title, description, meta."""
        return self._get_markdown_metadata()

    @property
    def markdown_metadata_filepath(self):
        return splitext(self.src_path)[0] + ".md"

    def _get_markdown_metadata(self):
        """Get metadata from filename.md."""
        meta = {"title": "", "description": "", "meta": {}}
        if isfile(self.markdown_metadata_filepath):
            meta.update(read_markdown(self.markdown_metadata_filepath))
        return meta

    @cached_property
    def file_metadata(self):
        """Type-specific metadata"""
        return {}

    def _get_file_date(self):
        return datetime.fromtimestamp(get_mod_date(self.src_path))


class Image(Media):
    """Gather all informations on an image file."""

    type = "image"

    def __init__(self, filename, path, settings):
        super().__init__(filename, path, settings)
        imgformat = settings.get("img_format")

        # Register all formats
        PILImage.init()

        if imgformat and PILImage.EXTENSION[self.src_ext] != imgformat.upper():
            # Find the extension that should match img_format
            extensions = {v: k for k, v in PILImage.EXTENSION.items()}
            ext = extensions[imgformat.upper()]
            self.dst_filename = self.basename + ext
            self.thumb_name = get_thumb(self.settings, self.dst_filename)

    @cached_property
    def date(self):
        """The date from the EXIF DateTimeOriginal metadata if available, or
        from the file date."""
        return self.exif and self.exif.get("dateobj", None) or self._get_file_date()

    @cached_property
    def exif(self):
        """If not `None` contains a dict with the most common tags. For more
        information, see :ref:`simple-exif-data`.
        """
        datetime_format = self.settings["datetime_format"]
        return (
            get_exif_tags(self.raw_exif, datetime_format=datetime_format)
            if self.raw_exif and self.src_ext in (".jpg", ".jpeg")
            else None
        )

    @cached_property
    def file_metadata(self):
        """Image file metadata (Exif and IPTC)"""
        return get_image_metadata(self.src_path)

    def _get_markdown_metadata(self):
        """Get metadata from filename.md."""
        meta = super()._get_markdown_metadata()

        # If a title or description hasn't been obtained by other means, look
        #  for the information in IPTC fields
        if not meta["title"]:
            meta["title"] = self.file_metadata["iptc"].get("title", "")
        if not meta["description"]:
            meta["description"] = self.file_metadata["iptc"].get("description", "")

        return meta

    @cached_property
    def raw_exif(self):
        """If not `None`, contains the raw EXIF tags."""
        if self.src_ext in (".jpg", ".jpeg"):
            return self.file_metadata["exif"]

    @cached_property
    def size(self):
        """The dimensions of the resized image."""
        return get_size(self.dst_path)

    @cached_property
    def input_size(self):
        """The dimensions of the input image."""
        return get_size(self.src_path)

    @cached_property
    def thumb_size(self):
        """The dimensions of the thumbnail image."""
        return get_size(self.thumb_path)

    def has_location(self):
        """True if location information is available for EXIF GPSInfo."""
        return self.exif is not None and "gps" in self.exif


class Video(Media):
    """Gather all informations on a video file."""

    type = "video"

    def __init__(self, filename, path, settings):
        super().__init__(filename, path, settings)

        if not settings["use_orig"] or not is_valid_html5_video(self.src_ext):
            video_format = settings["video_format"]
            ext = "." + video_format
            self.dst_filename = self.basename + ext
            self.mime = get_mime(ext)
        else:
            self.mime = get_mime(self.src_ext)

    @cached_property
    def date(self):
        """The date from the Date metadata if available, or from the file date."""
        if "date" in self.meta:
            try:
                self.logger.debug(
                    "Reading date from image metadata : %s", self.src_filename
                )
                return datetime.fromisoformat(self.meta["date"][0])
            except Exception:
                self.logger.debug(
                    "Reading date from image metadata failed : %s", self.src_filename
                )
        # If no date is found in the metadata, return the file date.
        return self._get_file_date()


class Album:
    """Gather all informations on an album.

    Attributes:

    :var description_file: Name of the Markdown file which gives information
        on an album
    :var index_url: URL to the index page.
    :var output_file: Name of the output HTML file
    :var meta: Meta data from the Markdown file.
    :var description: description from the Markdown file.

    For details how to annotate your albums with meta data, see
    :doc:`album_information`.

    """

    description_file = "index.md"

    def __init__(self, path, settings, dirnames, filenames, gallery):
        self.path = path
        self.name = path.split(os.path.sep)[-1]
        self.gallery = gallery
        self.settings = settings
        self.subdirs = dirnames
        self.output_file = settings["output_filename"]
        self._thumbnail = None

        if path == ".":
            self.src_path = settings["source"]
            self.dst_path = settings["destination"]
        else:
            self.src_path = join(settings["source"], path)
            self.dst_path = join(settings["destination"], path)

        self.logger = logging.getLogger(__name__)

        # optionally add index.html to the URLs
        self.url_ext = self.output_file if settings["index_in_url"] else ""

        self.index_url = (
            url_from_path(os.path.relpath(settings["destination"], self.dst_path))
            + "/"
            + self.url_ext
        )

        #: List of all medias in the album (:class:`~sigal.gallery.Image` and
        #: :class:`~sigal.gallery.Video`).
        self.medias = medias = []
        self.medias_count = defaultdict(int)

        for f in filenames:
            ext = splitext(f)[1]
            media = None
            if ext.lower() in settings["img_extensions"]:
                media = Image(f, self.path, settings)
            elif ext.lower() in settings["video_extensions"]:
                media = Video(f, self.path, settings)

            # Allow modification of the media, including overriding the class
            # type for the media.
            result = signals.album_file.send(self, filename=f, media=media)
            for recv, ret in result:
                if ret is not None:
                    media = ret

            if media:
                self.medias_count[media.type] += 1
                medias.append(media)

        signals.album_initialized.send(self)

    def __repr__(self):
        return f"<{self.__class__.__name__}>(path={self.path!r}, title={self.title!r})"

    def __str__(self):
        return f"{self.path} : " + ", ".join(
            f"{count} {_type}s" for _type, count in self.medias_count.items()
        )

    def __len__(self):
        return len(self.medias)

    def __iter__(self):
        return iter(self.medias)

    @cached_property
    def description(self):
        """Description extracted from the Markdown index.md file."""
        return self.markdown_metadata.get("description", "")

    @cached_property
    def title(self):
        """Title extracted from the Markdown index.md file."""
        title = self.markdown_metadata.get("title", "")
        path = self.path if self.path != "." else self.src_path
        return title if title else os.path.basename(path)

    @cached_property
    def meta(self):
        """Other metadata extracted from the Markdown index.md file."""
        return self.markdown_metadata.get("meta", {})

    @cached_property
    def author(self):
        """Author extracted from the Markdown index.md file or settings."""
        try:
            return self.meta["author"][0]
        except KeyError:
            return self.settings.get("author")

    @property
    def markdown_metadata_filepath(self):
        return join(self.src_path, self.description_file)

    @cached_property
    def markdown_metadata(self):
        """Get metadata from filename.md: title, description, meta."""
        meta = {"title": "", "description": "", "meta": {}}
        if isfile(self.markdown_metadata_filepath):
            meta.update(read_markdown(self.markdown_metadata_filepath))
        return meta

    def create_output_directories(self):
        """Create output directories for thumbnails and original images."""
        check_or_create_dir(self.dst_path)

        if self.medias:
            check_or_create_dir(join(self.dst_path, self.settings["thumb_dir"]))

        if self.medias and self.settings["keep_orig"]:
            self.orig_path = join(self.dst_path, self.settings["orig_dir"])
            check_or_create_dir(self.orig_path)

    def sort_subdirs(self, albums_sort_attr):
        if self.subdirs:
            if not albums_sort_attr:
                albums_sort_attr = self.settings["albums_sort_attr"]
            reverse = self.settings["albums_sort_reverse"]

            if "sort" in self.meta:
                albums_sort_attr = self.meta["sort"][0]
                if albums_sort_attr[0] == "-":
                    albums_sort_attr = albums_sort_attr[1:]
                    reverse = True
                else:
                    reverse = False

            root_path = self.path if self.path != "." else ""

            def sort_key(s):
                sort_attr = albums_sort_attr
                if not isinstance(sort_attr, list):
                    sort_attr = [sort_attr]

                album = self.gallery.albums[join(root_path, s)]

                for k in sort_attr:
                    try:
                        if k.startswith("meta."):
                            meta_key = k.split(".", 1)[1]
                            return album.meta.get(meta_key)[0]
                        else:
                            return getattr(album, k)
                    except AttributeError:
                        continue
                    except TypeError:
                        continue
                return ""

            key = natsort_keygen(key=sort_key, alg=ns.LOCALE)
            self.subdirs.sort(key=key, reverse=reverse)

        signals.albums_sorted.send(self)

    def sort_medias(self, medias_sort_attr):
        if self.medias:
            if medias_sort_attr == "filename":
                medias_sort_attr = "dst_filename"

            if medias_sort_attr == "date":
                key = lambda s: s.date or datetime.now()
            elif medias_sort_attr.startswith("meta."):
                meta_key = medias_sort_attr.split(".", 1)[1]
                key = natsort_keygen(
                    key=lambda s: s.meta.get(meta_key, [""])[0], alg=ns.LOCALE
                )
            else:
                key = natsort_keygen(
                    key=lambda s: getattr(s, medias_sort_attr), alg=ns.LOCALE
                )

            self.medias.sort(key=key, reverse=self.settings["medias_sort_reverse"])

        signals.medias_sorted.send(self)

    @property
    def images(self):
        """List of images (:class:`~sigal.gallery.Image`)."""
        for media in self.medias:
            if media.type == "image":
                yield media

    @property
    def videos(self):
        """List of videos (:class:`~sigal.gallery.Video`)."""
        for media in self.medias:
            if media.type == "video":
                yield media

    @property
    def albums(self):
        """List of :class:`~sigal.gallery.Album` objects for each
        sub-directory.
        """
        root_path = self.path if self.path != "." else ""
        return [self.gallery.albums[join(root_path, path)] for path in self.subdirs]

    @property
    def nbmedias(self):
        return len(self.medias) + sum(len(album) for album in self.albums)

    @property
    def url(self):
        """URL of the album, relative to its parent."""
        url = self.name.encode("utf-8")
        return url_quote(url) + "/" + self.url_ext

    @property
    def thumbnail(self):
        """Path to the thumbnail of the album."""

        if self._thumbnail:
            # stop if it is already set
            return self._thumbnail

        # Test the thumbnail from the Markdown file.
        thumbnail = self.meta.get("thumbnail", [""])[0]

        if thumbnail and isfile(join(self.src_path, thumbnail)):
            self._thumbnail = url_from_path(
                join(self.name, get_thumb(self.settings, thumbnail))
            )
            self.logger.debug("Thumbnail for %r : %s", self, self._thumbnail)
            return self._thumbnail
        else:
            # find and return the first landscape image
            for f in self.medias:
                ext = splitext(f.dst_filename)[1]
                if ext.lower() not in self.settings["img_extensions"]:
                    continue

                # Use f.size if available as it is quicker (in cache), but
                # fallback to the size of src_path if dst_path is missing
                size = f.input_size
                if size is None:
                    size = f.file_metadata["size"]

                if size["width"] > size["height"]:
                    try:
                        self._thumbnail = url_quote(self.name) + "/" + f.thumbnail
                    except Exception as e:
                        self.logger.info(
                            "Failed to get thumbnail for %s: %s", f.dst_filename, e
                        )
                    else:
                        self.logger.debug(
                            "Use 1st landscape image as thumbnail for %r : %s",
                            self,
                            self._thumbnail,
                        )
                        return self._thumbnail

            # else simply return the 1st media file
            if not self._thumbnail and self.medias:
                for media in self.medias:
                    if media.thumbnail is not None:
                        try:
                            self._thumbnail = (
                                url_quote(self.name) + "/" + media.thumbnail
                            )
                        except Exception as e:
                            self.logger.info(
                                "Failed to get thumbnail for %s: %s",
                                media.dst_filename,
                                e,
                            )
                        else:
                            break
                else:
                    self.logger.warning("No thumbnail found for %r", self)
                    return

                self.logger.debug(
                    "Use the 1st image as thumbnail for %r : %s", self, self._thumbnail
                )
                return self._thumbnail

            # use the thumbnail of their sub-directories
            if not self._thumbnail:
                for path, album in self.gallery.get_albums(self.path):
                    if album.thumbnail:
                        self._thumbnail = url_quote(self.name) + "/" + album.thumbnail
                        self.logger.debug(
                            "Using thumbnail from sub-directory for %r : %s",
                            self,
                            self._thumbnail,
                        )
                        return self._thumbnail

        self.logger.error("Thumbnail not found for %r", self)

    @property
    def random_thumbnail(self):
        try:
            return url_from_path(join(self.name, random.choice(self.medias).thumbnail))
        except IndexError:
            return self.thumbnail

    @property
    def breadcrumb(self):
        """List of ``(url, title)`` tuples defining the current breadcrumb
        path.
        """
        if self.path == ".":
            return []

        path = self.path
        breadcrumb = [((self.url_ext or "."), self.title)]

        while True:
            path = os.path.normpath(os.path.join(path, ".."))
            if path == ".":
                break

            url = url_from_path(os.path.relpath(path, self.path)) + "/" + self.url_ext
            breadcrumb.append((url, self.gallery.albums[path].title))

        breadcrumb.reverse()
        return breadcrumb

    @property
    def show_map(self):
        """Check if we have at least one photo with GPS location in the album"""
        return any(image.has_location() for image in self.images)

    @cached_property
    def zip(self):
        """Placeholder ZIP method.
        The ZIP logic is controlled by the zip_gallery plugin
        """


class Gallery:
    def __init__(self, settings, ncpu=None, show_progress=False):
        self.settings = settings
        self.logger = logging.getLogger(__name__)
        self.stats = defaultdict(int)
        self.init_pool(ncpu)
        check_or_create_dir(settings["destination"])

        if settings["max_img_pixels"]:
            PILImage.MAX_IMAGE_PIXELS = settings["max_img_pixels"]

        # Build the list of directories with images
        albums = self.albums = {}
        src_path = self.settings["source"]

        ignore_dirs = settings["ignore_directories"]
        ignore_files = settings["ignore_files"]

        progressChars = cycle(["/", "-", "\\", "|"])
        try:
            isatty = os.isatty(sys.stdout.fileno())
        except io.UnsupportedOperation:
            isatty = False

        self.progressbar_target = None if show_progress and isatty else Devnull()

        for path, dirs, files in os.walk(src_path, followlinks=True, topdown=False):
            if show_progress:
                print("\rCollecting albums " + next(progressChars), end="")
            relpath = os.path.relpath(path, src_path)

            # Test if the directory match the ignore_dirs settings
            if ignore_dirs and any(
                fnmatch.fnmatch(relpath, ignore) for ignore in ignore_dirs
            ):
                self.logger.info("Ignoring %s", relpath)
                continue

            # Remove files that match the ignore_files settings
            if ignore_files:
                files_path = {join(relpath, f) for f in files}
                for ignore in ignore_files:
                    files_path -= set(fnmatch.filter(files_path, ignore))

                self.logger.debug("Files before filtering: %r", files)
                files = [os.path.split(f)[1] for f in files_path]
                self.logger.debug("Files after filtering: %r", files)

            # Remove sub-directories that have been ignored in a previous
            # iteration (as topdown=False, sub-directories are processed before
            # their parent
            for d in dirs[:]:
                path = join(relpath, d) if relpath != "." else d
                if path not in albums.keys():
                    dirs.remove(d)

            album = Album(relpath, settings, dirs, files, self)

            if not album.medias and not album.albums:
                self.logger.info("Skip empty album: %r", album)
            else:
                album.create_output_directories()
                albums[relpath] = album

        if show_progress:
            print("\rCollecting albums, done.")

        with progressbar(
            albums.values(),
            label="%16s" % "Sorting albums",
            file=self.progressbar_target,
        ) as progress_albums:
            for album in progress_albums:
                album.sort_subdirs(settings["albums_sort_attr"])

        with progressbar(
            albums.values(),
            label="%16s" % "Sorting media",
            file=self.progressbar_target,
        ) as progress_albums:
            for album in progress_albums:
                album.sort_medias(settings["medias_sort_attr"])

        self.logger.debug("Albums:\n%r", albums.values())
        signals.gallery_initialized.send(self)

    @property
    def title(self):
        """Title of the gallery."""
        return self.settings["title"] or self.albums["."].title

    def init_pool(self, ncpu):
        try:
            cpu_count = multiprocessing.cpu_count()
        except NotImplementedError:
            cpu_count = 1

        if ncpu is None:
            ncpu = cpu_count
        else:
            try:
                ncpu = int(ncpu)
            except ValueError:
                self.logger.error("ncpu should be an integer value")
                ncpu = cpu_count

        self.logger.info("Using %s cores", ncpu)
        if ncpu > 1:
            self.pool = multiprocessing.Pool(
                processes=ncpu,
                initializer=pool_init,
                initargs=(self.settings["max_img_pixels"],),
            )
        else:
            self.pool = None

    def get_albums(self, path):
        """Return the list of all sub-directories of path."""

        for name in self.albums[path].subdirs:
            subdir = os.path.normpath(join(path, name))
            yield subdir, self.albums[subdir]
            for subname, album in self.get_albums(subdir):
                yield subname, self.albums[subdir]

    def build(self, force=False):
        "Create the image gallery"

        if not self.albums:
            self.logger.warning("No albums found.")
            return

        def log_func(x):
            # 63 is the total length of progressbar, label, percentage, etc
            available_length = get_terminal_size()[0] - 64
            if x and available_length > 10:
                return x.name[:available_length]
            else:
                return ""

        try:
            with progressbar(
                self.albums.values(),
                label="Collecting files",
                item_show_func=log_func,
                show_eta=False,
                file=self.progressbar_target,
            ) as albums:
                media_list = [
                    f for album in albums for f in self.process_dir(album, force=force)
                ]
        except KeyboardInterrupt:
            sys.exit("Interrupted")

        bar_opt = {
            "label": "Processing files",
            "show_pos": True,
            "file": self.progressbar_target,
        }

        if self.pool:
            result = []
            try:
                with progressbar(length=len(media_list), **bar_opt) as bar:
                    for status in self.pool.imap_unordered(worker, media_list):
                        result.append(status)
                        bar.update(1)
            except KeyboardInterrupt:
                self.pool.terminate()
                sys.exit("Interrupted")
            except pickle.PicklingError:
                self.logger.critical(
                    "Failed to process files with the multiprocessing feature."
                    " This can be caused by some module import or object "
                    "defined in the settings file, which can't be serialized.",
                    exc_info=True,
                )
                sys.exit("Abort")
            finally:
                self.pool.close()
                self.pool.join()
        else:
            with progressbar(media_list, **bar_opt) as medias:
                result = [process_file(media_item) for media_item in medias]

        if any(result):
            failed_files = [
                media for status, media in zip(result, media_list) if status != 0
            ]
            self.remove_files(failed_files)

        if self.settings["write_html"]:
            album_writer = AlbumPageWriter(self.settings, index_title=self.title)
            album_list_writer = AlbumListPageWriter(
                self.settings, index_title=self.title
            )
            with progressbar(
                self.albums.values(),
                label="%16s" % "Writing files",
                item_show_func=log_func,
                show_eta=False,
                file=self.progressbar_target,
            ) as albums:
                for album in albums:
                    if album.albums:
                        if album.medias:
                            self.logger.warning(
                                "Album '%s' contains sub-albums and images. "
                                "Please move images to their own sub-album. "
                                "Images in album %s will not be visible.",
                                album.title,
                                album.title,
                            )
                        album_list_writer.write(album)
                    else:
                        album_writer.write(album)
        print("")

        signals.gallery_build.send(self)

    def remove_files(self, medias):
        self.logger.error("Some files have failed to be processed:")
        for media in medias:
            self.logger.error("  - %s", media.dst_filename)
            album = self.albums[media.path]
            for f in album.medias:
                if f.dst_filename == media.dst_filename:
                    self.stats[f.type + "_failed"] += 1
                    album.medias.remove(f)
                    break
        self.logger.error(
            'You can run "sigal build" in verbose (--verbose) or'
            " debug (--debug) mode to get more details."
        )

    def process_dir(self, album, force=False):
        """Process a list of images in a directory."""
        for f in album:
            if isfile(f.dst_path) and not should_reprocess_album(
                album.path, album.name, force
            ):
                self.logger.info("%s exists - skipping", f.dst_filename)
                self.stats[f.type + "_skipped"] += 1
            else:
                self.stats[f.type] += 1
                yield f


def pool_init(max_img_pixels):
    if max_img_pixels:
        PILImage.MAX_IMAGE_PIXELS = max_img_pixels


def process_file(media):
    processor = None
    if media.type == "image":
        processor = process_image
    elif media.type == "video":
        processor = process_video

    # Allow overriding of the processor
    result = signals.process_file.send(media, processor=processor)
    for recv, ret in result:
        if ret is not None:
            processor = ret

    if processor:
        return processor(media)
    else:
        logging.warning("Processor not found for media %s", media.path)
        return Status.FAILURE


def worker(args):
    try:
        return process_file(args)
    except KeyboardInterrupt:
        pass



"""Plugin to upload generated files to Amazon S3.

This plugin requires boto_. All generated files are uploaded to a specified S3
bucket.  When using this plugin you have to make sure that the bucket already
exists and the you have access to the S3 bucket. The access credentials are
managed by boto_ and can be given as environment variables, configuration files
etc. More information can be found on the boto_ documentation.

.. _boto: https://pypi.org/project/boto/

Settings (all settings are wrapped in ``upload_s3_options`` dict):

- ``bucket``: The to-be-used bucket for uploading.
- ``policy``: Specifying access control to the uploaded files. Possible values:
  private, public-read, public-read-write, authenticated-read
- ``overwrite``: Boolean indicating if all files should be uploaded and
  overwritten or if already uploaded files should be skipped.
- ``max_age``: Optional, Integer indicating the number of seconds that the
  cache control should be set by default
- ``media_max_age``: Optional, Integer indicates the number of seconds that
  cache control hould be set for media files

"""

import logging
import os

from click import progressbar

from sigal import signals

logger = logging.getLogger(__name__)


def upload_s3(gallery, settings=None):
    import boto

    upload_files = []

    # Get local files
    for root, dirs, files in os.walk(gallery.settings["destination"]):
        for f in files:
            path = os.path.join(root[len(gallery.settings["destination"]) + 1 :], f)
            size = os.path.getsize(os.path.join(root, f))
            upload_files += [(path, size)]

    # Connect to specified bucket
    conn = boto.connect_s3()
    bucket = conn.get_bucket(gallery.settings["upload_s3_options"]["bucket"])

    # Upload the files
    with progressbar(upload_files, label="Uploading files to S3") as bar:
        for f, size in bar:
            if gallery.settings["upload_s3_options"]["overwrite"] is False:
                # Check if file was uploaded before
                key = bucket.get_key(f)
                if key is not None and key.size == size:
                    cache_metadata = generate_cache_metadata(gallery, f)

                    if key.get_metadata("Cache-Control") != cache_metadata:
                        key.set_remote_metadata(
                            {"Cache-Control": cache_metadata}, {}, True
                        )
                    logger.debug("Skipping file %s", f)
                else:
                    upload_file(gallery, bucket, f)
            else:
                # File is not available on S3 yet
                upload_file(gallery, bucket, f)


def generate_cache_metadata(gallery, f):
    filename, ext = os.path.splitext(f)
    options = gallery.settings["upload_s3_options"]

    proposed_cache_control = None
    if "media_max_age" in options and ext in [".jpg", ".png", ".webm", ".mp4"]:
        proposed_cache_control = "max-age=%s" % options["media_max_age"]
    elif "max_age" in options:
        proposed_cache_control = "max-age=%s" % options["max_age"]
    return proposed_cache_control


def upload_file(gallery, bucket, f):
    logger.debug("Uploading file %s", f)

    from boto.s3.key import Key

    key = Key(bucket)
    key.key = f

    cache_metadata = generate_cache_metadata(gallery, f)
    if cache_metadata:
        key.set_metadata("Cache-Control", cache_metadata)

    key.set_contents_from_filename(
        os.path.join(gallery.settings["destination"], f),
        policy=gallery.settings["upload_s3_options"]["policy"],
    )


def register(settings):
    if settings.get("upload_s3_options"):
        signals.gallery_build.connect(upload_s3)
    else:
        logger.warning("Upload to S3 is not configured.")



# Copyright (c) 2013-2023 - Simon Conseil

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import logging
import os
import sys
from logging import Formatter

# The background is set with 40 plus the number of the color, and the
# foreground with 30
BLACK, RED, GREEN, YELLOW, BLUE, MAGENTA, CYAN, WHITE = (30 + i for i in range(8))

COLORS = {
    "DEBUG": BLUE,
    "INFO": GREEN,
    "WARNING": YELLOW,
    "ERROR": RED,
    "CRITICAL": MAGENTA,
}

# These are the sequences need to get colored ouput
RESET_SEQ = "\033[0m"
COLOR_SEQ = "\033[1;{0}m{1}{2}"
BOLD_SEQ = "\033[1m"


def colored(text, color):
    return COLOR_SEQ.format(color, text, RESET_SEQ)


class ColoredFormatter(Formatter):
    def format(self, record):
        level = record.levelname
        return colored(level, COLORS[level]) + ": " + record.getMessage()


def init_logging(name, level=logging.INFO):
    """Logging config

    Set the level and create a more detailed formatter for debug mode.

    """
    logger = logging.getLogger(name)
    logger.setLevel(level)

    try:
        if os.isatty(sys.stdout.fileno()) and not sys.platform.startswith("win"):
            formatter = ColoredFormatter()
        elif level == logging.DEBUG:
            formatter = Formatter("%(levelname)s - %(message)s")
        else:
            formatter = Formatter("%(message)s")
    except Exception:
        # This fails when running tests with click (test_build)
        formatter = Formatter("%(message)s")

    handler = logging.StreamHandler()
    handler.setFormatter(formatter)
    logger.addHandler(handler)



"""Plugin which add RSS/ATOM feeds.

This plugin requires feedgenerator_. It uses all the images and videos of the
gallery, sorted by date, to show the most recent ones.

.. _feedgenerator: https://pypi.org/project/feedgenerator/

Settings:

- ``rss_feed`` and ``atom_feed``, see below.

Example::

    rss_feed = {'feed_url': 'http://example.org/feed.rss', 'nb_items': 10}
    atom_feed = {'feed_url': 'http://example.org/feed.atom', 'nb_items': 10}

"""

import logging
import os
from datetime import datetime

from markupsafe import Markup

from sigal import signals

logger = logging.getLogger(__name__)


def generate_feeds(gallery):
    # Get all images and videos and sort by date
    medias = [
        med
        for album in gallery.albums.values()
        for med in album.medias
        if med.date is not None
    ]
    medias.sort(key=lambda m: m.date, reverse=True)

    settings = gallery.settings
    if settings.get("rss_feed"):
        generate_feed(gallery, medias, feed_type="rss", **settings["rss_feed"])
    if settings.get("atom_feed"):
        generate_feed(gallery, medias, feed_type="atom", **settings["atom_feed"])


def generate_feed(gallery, medias, feed_type=None, feed_url="", nb_items=0):
    from feedgenerator import Atom1Feed, Rss201rev2Feed

    root_album = gallery.albums["."]
    cls = Rss201rev2Feed if feed_type == "rss" else Atom1Feed
    feed = cls(
        title=Markup.escape(root_album.title),
        link="/",
        feed_url=feed_url,
        description=Markup.escape(root_album.description).striptags(),
    )

    theme = gallery.settings["theme"]
    nb_medias = len(medias)
    nb_items = min(nb_items, nb_medias) if nb_items > 0 else nb_medias
    base_url = feed_url.rsplit("/", maxsplit=1)[0]

    for item in medias[:nb_items]:
        if theme == "galleria":
            link = f"{base_url}/{item.path}/#{item.url}"
        else:
            link = f"{base_url}/{item.path}/"

        feed.add_item(
            title=Markup.escape(item.title or item.url),
            link=link,
            # unique_id='tag:%s,%s:%s' % (urlparse(link).netloc,
            #                             item.date.date(),
            #                             urlparse(link).path.lstrip('/')),
            description=f'<img src="{base_url}/{item.path}/{item.thumbnail}" />',
            # categories=item.tags if hasattr(item, 'tags') else None,
            author_name=getattr(item, "author", ""),
            pubdate=item.date or datetime.now(),
        )

    output_file = os.path.join(root_album.dst_path, feed_url.split("/")[-1])
    logger.info("Generate %s feeds: %s", feed_type.upper(), output_file)
    with open(output_file, "w", encoding="utf8") as f:
        feed.write(f, "utf-8")


def register(settings):
    signals.gallery_build.connect(generate_feeds)



# Copyright 2022 - C.Sehrt

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

""" This plugin modifies titles of galleries by using regular-expressions and
simple string or character replacements. It is acting in two phases: First, all
regular-expression-based modifications are carried out, second, string/character
replacements are done.

The first phase may be interrupted if a match occurs by the 'break'-setting
individually per regular-expression part. Additionally, if a match occurs,
string/character replacements may be added.

The second phase is done even if the first phase had been interrupted.

Settings:

- ``titleregexp`` with the following keys:
    - ``regexp``, which is an array of dicts with 'search', 'replace', 'break'
      'substitute' and 'count' keys. All but 'break' and 'substitute' are the
      arguments for ``re.subn``, without given 'count' all matches are replaced.
      If 'break' is anything but an empty string, the rest of the
      ``regexp``-array is being skipped. The 'substitute'-key contains an array
      following the ``substitute`` format explained in the next sentence and
      does string-replacement if the regular-expression matched.
    - ``substitute``, which is an array of 2-element-arrays, of which the
      occurences of the first element will be replaced by the second element by
      the ``replace``-method of strings.

Example::

    titleregexp = {
        'regexp' : [
            { 'search': r"^([0-9]*)-(.*)$", 'replace': r"\\2 (\\1)", 'count': 1,
              'break': 1, substitute: [ ['ae','ä'] ] },
            { 'search': r"([a-z][a-z])([A-Z][a-z])", 'replace': r"\\1 \\2" }
            ],
        'substitute' : [ [ '_', ' ' ] ]
    }

"""

import logging
import re

from sigal import signals

logger = logging.getLogger(__name__)


def titleregexp(album):
    """Create a title by regexping name"""

    cfg = album.settings.get("titleregexp")
    n = 0
    total = 0
    album_title_org = album.title

    for r in cfg.get("regexp"):
        album.title, n = re.subn(
            r.get("search"), r.get("replace"), album.title, r.get("count", 0)
        )
        total += n

        if n > 0:
            for s in r.get("substitute", []):
                album.title = album.title.replace(s[0], s[1])
            if r.get("break", "") != "":
                break

    for r in cfg.get("substitute", []):
        album.title = album.title.replace(r[0], r[1])

    if total > 0:
        logger.info("Fixing title '%s' to '%s'", album_title_org, album.title)


def register(settings):
    if settings.get("titleregexp"):
        signals.album_initialized.connect(titleregexp)
    else:
        logger.warning("'titleregexp' setting not available!")
